# Generated from kernSpineParser.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .kernSpineParser import kernSpineParser
else:
    from kernSpineParser import kernSpineParser

# This class defines a complete listener for a parse tree produced by kernSpineParser.
class kernSpineParserListener(ParseTreeListener):

    # Enter a parse tree produced by kernSpineParser#start.
    def enterStart(self, ctx:kernSpineParser.StartContext):
        pass

    # Exit a parse tree produced by kernSpineParser#start.
    def exitStart(self, ctx:kernSpineParser.StartContext):
        pass


    # Enter a parse tree produced by kernSpineParser#field.
    def enterField(self, ctx:kernSpineParser.FieldContext):
        pass

    # Exit a parse tree produced by kernSpineParser#field.
    def exitField(self, ctx:kernSpineParser.FieldContext):
        pass


    # Enter a parse tree produced by kernSpineParser#notes_rests_chords.
    def enterNotes_rests_chords(self, ctx:kernSpineParser.Notes_rests_chordsContext):
        pass

    # Exit a parse tree produced by kernSpineParser#notes_rests_chords.
    def exitNotes_rests_chords(self, ctx:kernSpineParser.Notes_rests_chordsContext):
        pass


    # Enter a parse tree produced by kernSpineParser#structural.
    def enterStructural(self, ctx:kernSpineParser.StructuralContext):
        pass

    # Exit a parse tree produced by kernSpineParser#structural.
    def exitStructural(self, ctx:kernSpineParser.StructuralContext):
        pass


    # Enter a parse tree produced by kernSpineParser#contextual.
    def enterContextual(self, ctx:kernSpineParser.ContextualContext):
        pass

    # Exit a parse tree produced by kernSpineParser#contextual.
    def exitContextual(self, ctx:kernSpineParser.ContextualContext):
        pass


    # Enter a parse tree produced by kernSpineParser#signatures.
    def enterSignatures(self, ctx:kernSpineParser.SignaturesContext):
        pass

    # Exit a parse tree produced by kernSpineParser#signatures.
    def exitSignatures(self, ctx:kernSpineParser.SignaturesContext):
        pass


    # Enter a parse tree produced by kernSpineParser#otherContextual.
    def enterOtherContextual(self, ctx:kernSpineParser.OtherContextualContext):
        pass

    # Exit a parse tree produced by kernSpineParser#otherContextual.
    def exitOtherContextual(self, ctx:kernSpineParser.OtherContextualContext):
        pass


    # Enter a parse tree produced by kernSpineParser#empty.
    def enterEmpty(self, ctx:kernSpineParser.EmptyContext):
        pass

    # Exit a parse tree produced by kernSpineParser#empty.
    def exitEmpty(self, ctx:kernSpineParser.EmptyContext):
        pass


    # Enter a parse tree produced by kernSpineParser#rest.
    def enterRest(self, ctx:kernSpineParser.RestContext):
        pass

    # Exit a parse tree produced by kernSpineParser#rest.
    def exitRest(self, ctx:kernSpineParser.RestContext):
        pass


    # Enter a parse tree produced by kernSpineParser#restChar_r.
    def enterRestChar_r(self, ctx:kernSpineParser.RestChar_rContext):
        pass

    # Exit a parse tree produced by kernSpineParser#restChar_r.
    def exitRestChar_r(self, ctx:kernSpineParser.RestChar_rContext):
        pass


    # Enter a parse tree produced by kernSpineParser#restDecoration.
    def enterRestDecoration(self, ctx:kernSpineParser.RestDecorationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#restDecoration.
    def exitRestDecoration(self, ctx:kernSpineParser.RestDecorationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#chord.
    def enterChord(self, ctx:kernSpineParser.ChordContext):
        pass

    # Exit a parse tree produced by kernSpineParser#chord.
    def exitChord(self, ctx:kernSpineParser.ChordContext):
        pass


    # Enter a parse tree produced by kernSpineParser#note.
    def enterNote(self, ctx:kernSpineParser.NoteContext):
        pass

    # Exit a parse tree produced by kernSpineParser#note.
    def exitNote(self, ctx:kernSpineParser.NoteContext):
        pass


    # Enter a parse tree produced by kernSpineParser#nonVisualTandemInterpretation.
    def enterNonVisualTandemInterpretation(self, ctx:kernSpineParser.NonVisualTandemInterpretationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#nonVisualTandemInterpretation.
    def exitNonVisualTandemInterpretation(self, ctx:kernSpineParser.NonVisualTandemInterpretationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#boundingBox.
    def enterBoundingBox(self, ctx:kernSpineParser.BoundingBoxContext):
        pass

    # Exit a parse tree produced by kernSpineParser#boundingBox.
    def exitBoundingBox(self, ctx:kernSpineParser.BoundingBoxContext):
        pass


    # Enter a parse tree produced by kernSpineParser#xywh.
    def enterXywh(self, ctx:kernSpineParser.XywhContext):
        pass

    # Exit a parse tree produced by kernSpineParser#xywh.
    def exitXywh(self, ctx:kernSpineParser.XywhContext):
        pass


    # Enter a parse tree produced by kernSpineParser#x.
    def enterX(self, ctx:kernSpineParser.XContext):
        pass

    # Exit a parse tree produced by kernSpineParser#x.
    def exitX(self, ctx:kernSpineParser.XContext):
        pass


    # Enter a parse tree produced by kernSpineParser#y.
    def enterY(self, ctx:kernSpineParser.YContext):
        pass

    # Exit a parse tree produced by kernSpineParser#y.
    def exitY(self, ctx:kernSpineParser.YContext):
        pass


    # Enter a parse tree produced by kernSpineParser#w.
    def enterW(self, ctx:kernSpineParser.WContext):
        pass

    # Exit a parse tree produced by kernSpineParser#w.
    def exitW(self, ctx:kernSpineParser.WContext):
        pass


    # Enter a parse tree produced by kernSpineParser#h.
    def enterH(self, ctx:kernSpineParser.HContext):
        pass

    # Exit a parse tree produced by kernSpineParser#h.
    def exitH(self, ctx:kernSpineParser.HContext):
        pass


    # Enter a parse tree produced by kernSpineParser#pageNumber.
    def enterPageNumber(self, ctx:kernSpineParser.PageNumberContext):
        pass

    # Exit a parse tree produced by kernSpineParser#pageNumber.
    def exitPageNumber(self, ctx:kernSpineParser.PageNumberContext):
        pass


    # Enter a parse tree produced by kernSpineParser#visualTandemInterpretation.
    def enterVisualTandemInterpretation(self, ctx:kernSpineParser.VisualTandemInterpretationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#visualTandemInterpretation.
    def exitVisualTandemInterpretation(self, ctx:kernSpineParser.VisualTandemInterpretationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#associatedIDS.
    def enterAssociatedIDS(self, ctx:kernSpineParser.AssociatedIDSContext):
        pass

    # Exit a parse tree produced by kernSpineParser#associatedIDS.
    def exitAssociatedIDS(self, ctx:kernSpineParser.AssociatedIDSContext):
        pass


    # Enter a parse tree produced by kernSpineParser#placeHolder.
    def enterPlaceHolder(self, ctx:kernSpineParser.PlaceHolderContext):
        pass

    # Exit a parse tree produced by kernSpineParser#placeHolder.
    def exitPlaceHolder(self, ctx:kernSpineParser.PlaceHolderContext):
        pass


    # Enter a parse tree produced by kernSpineParser#octaveShift.
    def enterOctaveShift(self, ctx:kernSpineParser.OctaveShiftContext):
        pass

    # Exit a parse tree produced by kernSpineParser#octaveShift.
    def exitOctaveShift(self, ctx:kernSpineParser.OctaveShiftContext):
        pass


    # Enter a parse tree produced by kernSpineParser#pianoHand.
    def enterPianoHand(self, ctx:kernSpineParser.PianoHandContext):
        pass

    # Exit a parse tree produced by kernSpineParser#pianoHand.
    def exitPianoHand(self, ctx:kernSpineParser.PianoHandContext):
        pass


    # Enter a parse tree produced by kernSpineParser#tandemTuplet.
    def enterTandemTuplet(self, ctx:kernSpineParser.TandemTupletContext):
        pass

    # Exit a parse tree produced by kernSpineParser#tandemTuplet.
    def exitTandemTuplet(self, ctx:kernSpineParser.TandemTupletContext):
        pass


    # Enter a parse tree produced by kernSpineParser#tandemCue.
    def enterTandemCue(self, ctx:kernSpineParser.TandemCueContext):
        pass

    # Exit a parse tree produced by kernSpineParser#tandemCue.
    def exitTandemCue(self, ctx:kernSpineParser.TandemCueContext):
        pass


    # Enter a parse tree produced by kernSpineParser#tandemTremolo.
    def enterTandemTremolo(self, ctx:kernSpineParser.TandemTremoloContext):
        pass

    # Exit a parse tree produced by kernSpineParser#tandemTremolo.
    def exitTandemTremolo(self, ctx:kernSpineParser.TandemTremoloContext):
        pass


    # Enter a parse tree produced by kernSpineParser#ossia.
    def enterOssia(self, ctx:kernSpineParser.OssiaContext):
        pass

    # Exit a parse tree produced by kernSpineParser#ossia.
    def exitOssia(self, ctx:kernSpineParser.OssiaContext):
        pass


    # Enter a parse tree produced by kernSpineParser#rscale.
    def enterRscale(self, ctx:kernSpineParser.RscaleContext):
        pass

    # Exit a parse tree produced by kernSpineParser#rscale.
    def exitRscale(self, ctx:kernSpineParser.RscaleContext):
        pass


    # Enter a parse tree produced by kernSpineParser#pedal.
    def enterPedal(self, ctx:kernSpineParser.PedalContext):
        pass

    # Exit a parse tree produced by kernSpineParser#pedal.
    def exitPedal(self, ctx:kernSpineParser.PedalContext):
        pass


    # Enter a parse tree produced by kernSpineParser#ela.
    def enterEla(self, ctx:kernSpineParser.ElaContext):
        pass

    # Exit a parse tree produced by kernSpineParser#ela.
    def exitEla(self, ctx:kernSpineParser.ElaContext):
        pass


    # Enter a parse tree produced by kernSpineParser#dynamics_position.
    def enterDynamics_position(self, ctx:kernSpineParser.Dynamics_positionContext):
        pass

    # Exit a parse tree produced by kernSpineParser#dynamics_position.
    def exitDynamics_position(self, ctx:kernSpineParser.Dynamics_positionContext):
        pass


    # Enter a parse tree produced by kernSpineParser#sections.
    def enterSections(self, ctx:kernSpineParser.SectionsContext):
        pass

    # Exit a parse tree produced by kernSpineParser#sections.
    def exitSections(self, ctx:kernSpineParser.SectionsContext):
        pass


    # Enter a parse tree produced by kernSpineParser#sectionNames.
    def enterSectionNames(self, ctx:kernSpineParser.SectionNamesContext):
        pass

    # Exit a parse tree produced by kernSpineParser#sectionNames.
    def exitSectionNames(self, ctx:kernSpineParser.SectionNamesContext):
        pass


    # Enter a parse tree produced by kernSpineParser#sectionName.
    def enterSectionName(self, ctx:kernSpineParser.SectionNameContext):
        pass

    # Exit a parse tree produced by kernSpineParser#sectionName.
    def exitSectionName(self, ctx:kernSpineParser.SectionNameContext):
        pass


    # Enter a parse tree produced by kernSpineParser#transposition.
    def enterTransposition(self, ctx:kernSpineParser.TranspositionContext):
        pass

    # Exit a parse tree produced by kernSpineParser#transposition.
    def exitTransposition(self, ctx:kernSpineParser.TranspositionContext):
        pass


    # Enter a parse tree produced by kernSpineParser#instrument.
    def enterInstrument(self, ctx:kernSpineParser.InstrumentContext):
        pass

    # Exit a parse tree produced by kernSpineParser#instrument.
    def exitInstrument(self, ctx:kernSpineParser.InstrumentContext):
        pass


    # Enter a parse tree produced by kernSpineParser#instrumentTitle.
    def enterInstrumentTitle(self, ctx:kernSpineParser.InstrumentTitleContext):
        pass

    # Exit a parse tree produced by kernSpineParser#instrumentTitle.
    def exitInstrumentTitle(self, ctx:kernSpineParser.InstrumentTitleContext):
        pass


    # Enter a parse tree produced by kernSpineParser#number.
    def enterNumber(self, ctx:kernSpineParser.NumberContext):
        pass

    # Exit a parse tree produced by kernSpineParser#number.
    def exitNumber(self, ctx:kernSpineParser.NumberContext):
        pass


    # Enter a parse tree produced by kernSpineParser#lowerCasePitch.
    def enterLowerCasePitch(self, ctx:kernSpineParser.LowerCasePitchContext):
        pass

    # Exit a parse tree produced by kernSpineParser#lowerCasePitch.
    def exitLowerCasePitch(self, ctx:kernSpineParser.LowerCasePitchContext):
        pass


    # Enter a parse tree produced by kernSpineParser#upperCasePitch.
    def enterUpperCasePitch(self, ctx:kernSpineParser.UpperCasePitchContext):
        pass

    # Exit a parse tree produced by kernSpineParser#upperCasePitch.
    def exitUpperCasePitch(self, ctx:kernSpineParser.UpperCasePitchContext):
        pass


    # Enter a parse tree produced by kernSpineParser#pitchClass.
    def enterPitchClass(self, ctx:kernSpineParser.PitchClassContext):
        pass

    # Exit a parse tree produced by kernSpineParser#pitchClass.
    def exitPitchClass(self, ctx:kernSpineParser.PitchClassContext):
        pass


    # Enter a parse tree produced by kernSpineParser#accomp.
    def enterAccomp(self, ctx:kernSpineParser.AccompContext):
        pass

    # Exit a parse tree produced by kernSpineParser#accomp.
    def exitAccomp(self, ctx:kernSpineParser.AccompContext):
        pass


    # Enter a parse tree produced by kernSpineParser#solo.
    def enterSolo(self, ctx:kernSpineParser.SoloContext):
        pass

    # Exit a parse tree produced by kernSpineParser#solo.
    def exitSolo(self, ctx:kernSpineParser.SoloContext):
        pass


    # Enter a parse tree produced by kernSpineParser#strophe.
    def enterStrophe(self, ctx:kernSpineParser.StropheContext):
        pass

    # Exit a parse tree produced by kernSpineParser#strophe.
    def exitStrophe(self, ctx:kernSpineParser.StropheContext):
        pass


    # Enter a parse tree produced by kernSpineParser#timebase.
    def enterTimebase(self, ctx:kernSpineParser.TimebaseContext):
        pass

    # Exit a parse tree produced by kernSpineParser#timebase.
    def exitTimebase(self, ctx:kernSpineParser.TimebaseContext):
        pass


    # Enter a parse tree produced by kernSpineParser#part.
    def enterPart(self, ctx:kernSpineParser.PartContext):
        pass

    # Exit a parse tree produced by kernSpineParser#part.
    def exitPart(self, ctx:kernSpineParser.PartContext):
        pass


    # Enter a parse tree produced by kernSpineParser#group.
    def enterGroup(self, ctx:kernSpineParser.GroupContext):
        pass

    # Exit a parse tree produced by kernSpineParser#group.
    def exitGroup(self, ctx:kernSpineParser.GroupContext):
        pass


    # Enter a parse tree produced by kernSpineParser#staff.
    def enterStaff(self, ctx:kernSpineParser.StaffContext):
        pass

    # Exit a parse tree produced by kernSpineParser#staff.
    def exitStaff(self, ctx:kernSpineParser.StaffContext):
        pass


    # Enter a parse tree produced by kernSpineParser#clef.
    def enterClef(self, ctx:kernSpineParser.ClefContext):
        pass

    # Exit a parse tree produced by kernSpineParser#clef.
    def exitClef(self, ctx:kernSpineParser.ClefContext):
        pass


    # Enter a parse tree produced by kernSpineParser#clefValue.
    def enterClefValue(self, ctx:kernSpineParser.ClefValueContext):
        pass

    # Exit a parse tree produced by kernSpineParser#clefValue.
    def exitClefValue(self, ctx:kernSpineParser.ClefValueContext):
        pass


    # Enter a parse tree produced by kernSpineParser#clefSign.
    def enterClefSign(self, ctx:kernSpineParser.ClefSignContext):
        pass

    # Exit a parse tree produced by kernSpineParser#clefSign.
    def exitClefSign(self, ctx:kernSpineParser.ClefSignContext):
        pass


    # Enter a parse tree produced by kernSpineParser#clefLine.
    def enterClefLine(self, ctx:kernSpineParser.ClefLineContext):
        pass

    # Exit a parse tree produced by kernSpineParser#clefLine.
    def exitClefLine(self, ctx:kernSpineParser.ClefLineContext):
        pass


    # Enter a parse tree produced by kernSpineParser#clefOctave.
    def enterClefOctave(self, ctx:kernSpineParser.ClefOctaveContext):
        pass

    # Exit a parse tree produced by kernSpineParser#clefOctave.
    def exitClefOctave(self, ctx:kernSpineParser.ClefOctaveContext):
        pass


    # Enter a parse tree produced by kernSpineParser#keySignature.
    def enterKeySignature(self, ctx:kernSpineParser.KeySignatureContext):
        pass

    # Exit a parse tree produced by kernSpineParser#keySignature.
    def exitKeySignature(self, ctx:kernSpineParser.KeySignatureContext):
        pass


    # Enter a parse tree produced by kernSpineParser#keySignaturePitchClass.
    def enterKeySignaturePitchClass(self, ctx:kernSpineParser.KeySignaturePitchClassContext):
        pass

    # Exit a parse tree produced by kernSpineParser#keySignaturePitchClass.
    def exitKeySignaturePitchClass(self, ctx:kernSpineParser.KeySignaturePitchClassContext):
        pass


    # Enter a parse tree produced by kernSpineParser#keySignatureCancel.
    def enterKeySignatureCancel(self, ctx:kernSpineParser.KeySignatureCancelContext):
        pass

    # Exit a parse tree produced by kernSpineParser#keySignatureCancel.
    def exitKeySignatureCancel(self, ctx:kernSpineParser.KeySignatureCancelContext):
        pass


    # Enter a parse tree produced by kernSpineParser#keyCancel.
    def enterKeyCancel(self, ctx:kernSpineParser.KeyCancelContext):
        pass

    # Exit a parse tree produced by kernSpineParser#keyCancel.
    def exitKeyCancel(self, ctx:kernSpineParser.KeyCancelContext):
        pass


    # Enter a parse tree produced by kernSpineParser#keyMode.
    def enterKeyMode(self, ctx:kernSpineParser.KeyModeContext):
        pass

    # Exit a parse tree produced by kernSpineParser#keyMode.
    def exitKeyMode(self, ctx:kernSpineParser.KeyModeContext):
        pass


    # Enter a parse tree produced by kernSpineParser#key.
    def enterKey(self, ctx:kernSpineParser.KeyContext):
        pass

    # Exit a parse tree produced by kernSpineParser#key.
    def exitKey(self, ctx:kernSpineParser.KeyContext):
        pass


    # Enter a parse tree produced by kernSpineParser#singleKey.
    def enterSingleKey(self, ctx:kernSpineParser.SingleKeyContext):
        pass

    # Exit a parse tree produced by kernSpineParser#singleKey.
    def exitSingleKey(self, ctx:kernSpineParser.SingleKeyContext):
        pass


    # Enter a parse tree produced by kernSpineParser#minorKey.
    def enterMinorKey(self, ctx:kernSpineParser.MinorKeyContext):
        pass

    # Exit a parse tree produced by kernSpineParser#minorKey.
    def exitMinorKey(self, ctx:kernSpineParser.MinorKeyContext):
        pass


    # Enter a parse tree produced by kernSpineParser#majorKey.
    def enterMajorKey(self, ctx:kernSpineParser.MajorKeyContext):
        pass

    # Exit a parse tree produced by kernSpineParser#majorKey.
    def exitMajorKey(self, ctx:kernSpineParser.MajorKeyContext):
        pass


    # Enter a parse tree produced by kernSpineParser#modal.
    def enterModal(self, ctx:kernSpineParser.ModalContext):
        pass

    # Exit a parse tree produced by kernSpineParser#modal.
    def exitModal(self, ctx:kernSpineParser.ModalContext):
        pass


    # Enter a parse tree produced by kernSpineParser#locrian.
    def enterLocrian(self, ctx:kernSpineParser.LocrianContext):
        pass

    # Exit a parse tree produced by kernSpineParser#locrian.
    def exitLocrian(self, ctx:kernSpineParser.LocrianContext):
        pass


    # Enter a parse tree produced by kernSpineParser#ionian.
    def enterIonian(self, ctx:kernSpineParser.IonianContext):
        pass

    # Exit a parse tree produced by kernSpineParser#ionian.
    def exitIonian(self, ctx:kernSpineParser.IonianContext):
        pass


    # Enter a parse tree produced by kernSpineParser#aeolian.
    def enterAeolian(self, ctx:kernSpineParser.AeolianContext):
        pass

    # Exit a parse tree produced by kernSpineParser#aeolian.
    def exitAeolian(self, ctx:kernSpineParser.AeolianContext):
        pass


    # Enter a parse tree produced by kernSpineParser#mixolydian.
    def enterMixolydian(self, ctx:kernSpineParser.MixolydianContext):
        pass

    # Exit a parse tree produced by kernSpineParser#mixolydian.
    def exitMixolydian(self, ctx:kernSpineParser.MixolydianContext):
        pass


    # Enter a parse tree produced by kernSpineParser#lydian.
    def enterLydian(self, ctx:kernSpineParser.LydianContext):
        pass

    # Exit a parse tree produced by kernSpineParser#lydian.
    def exitLydian(self, ctx:kernSpineParser.LydianContext):
        pass


    # Enter a parse tree produced by kernSpineParser#phrygian.
    def enterPhrygian(self, ctx:kernSpineParser.PhrygianContext):
        pass

    # Exit a parse tree produced by kernSpineParser#phrygian.
    def exitPhrygian(self, ctx:kernSpineParser.PhrygianContext):
        pass


    # Enter a parse tree produced by kernSpineParser#dorian.
    def enterDorian(self, ctx:kernSpineParser.DorianContext):
        pass

    # Exit a parse tree produced by kernSpineParser#dorian.
    def exitDorian(self, ctx:kernSpineParser.DorianContext):
        pass


    # Enter a parse tree produced by kernSpineParser#timeSignature.
    def enterTimeSignature(self, ctx:kernSpineParser.TimeSignatureContext):
        pass

    # Exit a parse tree produced by kernSpineParser#timeSignature.
    def exitTimeSignature(self, ctx:kernSpineParser.TimeSignatureContext):
        pass


    # Enter a parse tree produced by kernSpineParser#numerator.
    def enterNumerator(self, ctx:kernSpineParser.NumeratorContext):
        pass

    # Exit a parse tree produced by kernSpineParser#numerator.
    def exitNumerator(self, ctx:kernSpineParser.NumeratorContext):
        pass


    # Enter a parse tree produced by kernSpineParser#denominator.
    def enterDenominator(self, ctx:kernSpineParser.DenominatorContext):
        pass

    # Exit a parse tree produced by kernSpineParser#denominator.
    def exitDenominator(self, ctx:kernSpineParser.DenominatorContext):
        pass


    # Enter a parse tree produced by kernSpineParser#standardTimeSignature.
    def enterStandardTimeSignature(self, ctx:kernSpineParser.StandardTimeSignatureContext):
        pass

    # Exit a parse tree produced by kernSpineParser#standardTimeSignature.
    def exitStandardTimeSignature(self, ctx:kernSpineParser.StandardTimeSignatureContext):
        pass


    # Enter a parse tree produced by kernSpineParser#additiveTimeSignature.
    def enterAdditiveTimeSignature(self, ctx:kernSpineParser.AdditiveTimeSignatureContext):
        pass

    # Exit a parse tree produced by kernSpineParser#additiveTimeSignature.
    def exitAdditiveTimeSignature(self, ctx:kernSpineParser.AdditiveTimeSignatureContext):
        pass


    # Enter a parse tree produced by kernSpineParser#mixedTimeSignature.
    def enterMixedTimeSignature(self, ctx:kernSpineParser.MixedTimeSignatureContext):
        pass

    # Exit a parse tree produced by kernSpineParser#mixedTimeSignature.
    def exitMixedTimeSignature(self, ctx:kernSpineParser.MixedTimeSignatureContext):
        pass


    # Enter a parse tree produced by kernSpineParser#alternatingTimeSignature.
    def enterAlternatingTimeSignature(self, ctx:kernSpineParser.AlternatingTimeSignatureContext):
        pass

    # Exit a parse tree produced by kernSpineParser#alternatingTimeSignature.
    def exitAlternatingTimeSignature(self, ctx:kernSpineParser.AlternatingTimeSignatureContext):
        pass


    # Enter a parse tree produced by kernSpineParser#alternatingTimeSignatureItem.
    def enterAlternatingTimeSignatureItem(self, ctx:kernSpineParser.AlternatingTimeSignatureItemContext):
        pass

    # Exit a parse tree produced by kernSpineParser#alternatingTimeSignatureItem.
    def exitAlternatingTimeSignatureItem(self, ctx:kernSpineParser.AlternatingTimeSignatureItemContext):
        pass


    # Enter a parse tree produced by kernSpineParser#interchangingTimeSignature.
    def enterInterchangingTimeSignature(self, ctx:kernSpineParser.InterchangingTimeSignatureContext):
        pass

    # Exit a parse tree produced by kernSpineParser#interchangingTimeSignature.
    def exitInterchangingTimeSignature(self, ctx:kernSpineParser.InterchangingTimeSignatureContext):
        pass


    # Enter a parse tree produced by kernSpineParser#meterSymbol.
    def enterMeterSymbol(self, ctx:kernSpineParser.MeterSymbolContext):
        pass

    # Exit a parse tree produced by kernSpineParser#meterSymbol.
    def exitMeterSymbol(self, ctx:kernSpineParser.MeterSymbolContext):
        pass


    # Enter a parse tree produced by kernSpineParser#modernMeterSymbolSign.
    def enterModernMeterSymbolSign(self, ctx:kernSpineParser.ModernMeterSymbolSignContext):
        pass

    # Exit a parse tree produced by kernSpineParser#modernMeterSymbolSign.
    def exitModernMeterSymbolSign(self, ctx:kernSpineParser.ModernMeterSymbolSignContext):
        pass


    # Enter a parse tree produced by kernSpineParser#mensuration.
    def enterMensuration(self, ctx:kernSpineParser.MensurationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#mensuration.
    def exitMensuration(self, ctx:kernSpineParser.MensurationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#metronome.
    def enterMetronome(self, ctx:kernSpineParser.MetronomeContext):
        pass

    # Exit a parse tree produced by kernSpineParser#metronome.
    def exitMetronome(self, ctx:kernSpineParser.MetronomeContext):
        pass


    # Enter a parse tree produced by kernSpineParser#nullInterpretation.
    def enterNullInterpretation(self, ctx:kernSpineParser.NullInterpretationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#nullInterpretation.
    def exitNullInterpretation(self, ctx:kernSpineParser.NullInterpretationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#barline.
    def enterBarline(self, ctx:kernSpineParser.BarlineContext):
        pass

    # Exit a parse tree produced by kernSpineParser#barline.
    def exitBarline(self, ctx:kernSpineParser.BarlineContext):
        pass


    # Enter a parse tree produced by kernSpineParser#barLineType.
    def enterBarLineType(self, ctx:kernSpineParser.BarLineTypeContext):
        pass

    # Exit a parse tree produced by kernSpineParser#barLineType.
    def exitBarLineType(self, ctx:kernSpineParser.BarLineTypeContext):
        pass


    # Enter a parse tree produced by kernSpineParser#restPosition.
    def enterRestPosition(self, ctx:kernSpineParser.RestPositionContext):
        pass

    # Exit a parse tree produced by kernSpineParser#restPosition.
    def exitRestPosition(self, ctx:kernSpineParser.RestPositionContext):
        pass


    # Enter a parse tree produced by kernSpineParser#duration.
    def enterDuration(self, ctx:kernSpineParser.DurationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#duration.
    def exitDuration(self, ctx:kernSpineParser.DurationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#fermata.
    def enterFermata(self, ctx:kernSpineParser.FermataContext):
        pass

    # Exit a parse tree produced by kernSpineParser#fermata.
    def exitFermata(self, ctx:kernSpineParser.FermataContext):
        pass


    # Enter a parse tree produced by kernSpineParser#modernDuration.
    def enterModernDuration(self, ctx:kernSpineParser.ModernDurationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#modernDuration.
    def exitModernDuration(self, ctx:kernSpineParser.ModernDurationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#augmentationDot.
    def enterAugmentationDot(self, ctx:kernSpineParser.AugmentationDotContext):
        pass

    # Exit a parse tree produced by kernSpineParser#augmentationDot.
    def exitAugmentationDot(self, ctx:kernSpineParser.AugmentationDotContext):
        pass


    # Enter a parse tree produced by kernSpineParser#alteration.
    def enterAlteration(self, ctx:kernSpineParser.AlterationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#alteration.
    def exitAlteration(self, ctx:kernSpineParser.AlterationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#staffChange.
    def enterStaffChange(self, ctx:kernSpineParser.StaffChangeContext):
        pass

    # Exit a parse tree produced by kernSpineParser#staffChange.
    def exitStaffChange(self, ctx:kernSpineParser.StaffChangeContext):
        pass


    # Enter a parse tree produced by kernSpineParser#chordSpace.
    def enterChordSpace(self, ctx:kernSpineParser.ChordSpaceContext):
        pass

    # Exit a parse tree produced by kernSpineParser#chordSpace.
    def exitChordSpace(self, ctx:kernSpineParser.ChordSpaceContext):
        pass


    # Enter a parse tree produced by kernSpineParser#graceNote.
    def enterGraceNote(self, ctx:kernSpineParser.GraceNoteContext):
        pass

    # Exit a parse tree produced by kernSpineParser#graceNote.
    def exitGraceNote(self, ctx:kernSpineParser.GraceNoteContext):
        pass


    # Enter a parse tree produced by kernSpineParser#appoggiatura.
    def enterAppoggiatura(self, ctx:kernSpineParser.AppoggiaturaContext):
        pass

    # Exit a parse tree produced by kernSpineParser#appoggiatura.
    def exitAppoggiatura(self, ctx:kernSpineParser.AppoggiaturaContext):
        pass


    # Enter a parse tree produced by kernSpineParser#appoggiaturaMode.
    def enterAppoggiaturaMode(self, ctx:kernSpineParser.AppoggiaturaModeContext):
        pass

    # Exit a parse tree produced by kernSpineParser#appoggiaturaMode.
    def exitAppoggiaturaMode(self, ctx:kernSpineParser.AppoggiaturaModeContext):
        pass


    # Enter a parse tree produced by kernSpineParser#ligatureTie.
    def enterLigatureTie(self, ctx:kernSpineParser.LigatureTieContext):
        pass

    # Exit a parse tree produced by kernSpineParser#ligatureTie.
    def exitLigatureTie(self, ctx:kernSpineParser.LigatureTieContext):
        pass


    # Enter a parse tree produced by kernSpineParser#noteDecoration.
    def enterNoteDecoration(self, ctx:kernSpineParser.NoteDecorationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#noteDecoration.
    def exitNoteDecoration(self, ctx:kernSpineParser.NoteDecorationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#noteDecorationCharX.
    def enterNoteDecorationCharX(self, ctx:kernSpineParser.NoteDecorationCharXContext):
        pass

    # Exit a parse tree produced by kernSpineParser#noteDecorationCharX.
    def exitNoteDecorationCharX(self, ctx:kernSpineParser.NoteDecorationCharXContext):
        pass


    # Enter a parse tree produced by kernSpineParser#phrase.
    def enterPhrase(self, ctx:kernSpineParser.PhraseContext):
        pass

    # Exit a parse tree produced by kernSpineParser#phrase.
    def exitPhrase(self, ctx:kernSpineParser.PhraseContext):
        pass


    # Enter a parse tree produced by kernSpineParser#diatonicPitchAndOctave.
    def enterDiatonicPitchAndOctave(self, ctx:kernSpineParser.DiatonicPitchAndOctaveContext):
        pass

    # Exit a parse tree produced by kernSpineParser#diatonicPitchAndOctave.
    def exitDiatonicPitchAndOctave(self, ctx:kernSpineParser.DiatonicPitchAndOctaveContext):
        pass


    # Enter a parse tree produced by kernSpineParser#trebleNotes.
    def enterTrebleNotes(self, ctx:kernSpineParser.TrebleNotesContext):
        pass

    # Exit a parse tree produced by kernSpineParser#trebleNotes.
    def exitTrebleNotes(self, ctx:kernSpineParser.TrebleNotesContext):
        pass


    # Enter a parse tree produced by kernSpineParser#bassNotes.
    def enterBassNotes(self, ctx:kernSpineParser.BassNotesContext):
        pass

    # Exit a parse tree produced by kernSpineParser#bassNotes.
    def exitBassNotes(self, ctx:kernSpineParser.BassNotesContext):
        pass


    # Enter a parse tree produced by kernSpineParser#accidental.
    def enterAccidental(self, ctx:kernSpineParser.AccidentalContext):
        pass

    # Exit a parse tree produced by kernSpineParser#accidental.
    def exitAccidental(self, ctx:kernSpineParser.AccidentalContext):
        pass


    # Enter a parse tree produced by kernSpineParser#alterationDisplay.
    def enterAlterationDisplay(self, ctx:kernSpineParser.AlterationDisplayContext):
        pass

    # Exit a parse tree produced by kernSpineParser#alterationDisplay.
    def exitAlterationDisplay(self, ctx:kernSpineParser.AlterationDisplayContext):
        pass


    # Enter a parse tree produced by kernSpineParser#turn.
    def enterTurn(self, ctx:kernSpineParser.TurnContext):
        pass

    # Exit a parse tree produced by kernSpineParser#turn.
    def exitTurn(self, ctx:kernSpineParser.TurnContext):
        pass


    # Enter a parse tree produced by kernSpineParser#userAssignable.
    def enterUserAssignable(self, ctx:kernSpineParser.UserAssignableContext):
        pass

    # Exit a parse tree produced by kernSpineParser#userAssignable.
    def exitUserAssignable(self, ctx:kernSpineParser.UserAssignableContext):
        pass


    # Enter a parse tree produced by kernSpineParser#glissando.
    def enterGlissando(self, ctx:kernSpineParser.GlissandoContext):
        pass

    # Exit a parse tree produced by kernSpineParser#glissando.
    def exitGlissando(self, ctx:kernSpineParser.GlissandoContext):
        pass


    # Enter a parse tree produced by kernSpineParser#articulation.
    def enterArticulation(self, ctx:kernSpineParser.ArticulationContext):
        pass

    # Exit a parse tree produced by kernSpineParser#articulation.
    def exitArticulation(self, ctx:kernSpineParser.ArticulationContext):
        pass


    # Enter a parse tree produced by kernSpineParser#accent.
    def enterAccent(self, ctx:kernSpineParser.AccentContext):
        pass

    # Exit a parse tree produced by kernSpineParser#accent.
    def exitAccent(self, ctx:kernSpineParser.AccentContext):
        pass


    # Enter a parse tree produced by kernSpineParser#tenuto.
    def enterTenuto(self, ctx:kernSpineParser.TenutoContext):
        pass

    # Exit a parse tree produced by kernSpineParser#tenuto.
    def exitTenuto(self, ctx:kernSpineParser.TenutoContext):
        pass


    # Enter a parse tree produced by kernSpineParser#staccatissimo.
    def enterStaccatissimo(self, ctx:kernSpineParser.StaccatissimoContext):
        pass

    # Exit a parse tree produced by kernSpineParser#staccatissimo.
    def exitStaccatissimo(self, ctx:kernSpineParser.StaccatissimoContext):
        pass


    # Enter a parse tree produced by kernSpineParser#pizzicato.
    def enterPizzicato(self, ctx:kernSpineParser.PizzicatoContext):
        pass

    # Exit a parse tree produced by kernSpineParser#pizzicato.
    def exitPizzicato(self, ctx:kernSpineParser.PizzicatoContext):
        pass


    # Enter a parse tree produced by kernSpineParser#spiccato.
    def enterSpiccato(self, ctx:kernSpineParser.SpiccatoContext):
        pass

    # Exit a parse tree produced by kernSpineParser#spiccato.
    def exitSpiccato(self, ctx:kernSpineParser.SpiccatoContext):
        pass


    # Enter a parse tree produced by kernSpineParser#staccato.
    def enterStaccato(self, ctx:kernSpineParser.StaccatoContext):
        pass

    # Exit a parse tree produced by kernSpineParser#staccato.
    def exitStaccato(self, ctx:kernSpineParser.StaccatoContext):
        pass


    # Enter a parse tree produced by kernSpineParser#editorialIntervention.
    def enterEditorialIntervention(self, ctx:kernSpineParser.EditorialInterventionContext):
        pass

    # Exit a parse tree produced by kernSpineParser#editorialIntervention.
    def exitEditorialIntervention(self, ctx:kernSpineParser.EditorialInterventionContext):
        pass


    # Enter a parse tree produced by kernSpineParser#slurStart.
    def enterSlurStart(self, ctx:kernSpineParser.SlurStartContext):
        pass

    # Exit a parse tree produced by kernSpineParser#slurStart.
    def exitSlurStart(self, ctx:kernSpineParser.SlurStartContext):
        pass


    # Enter a parse tree produced by kernSpineParser#ligatureTieStart.
    def enterLigatureTieStart(self, ctx:kernSpineParser.LigatureTieStartContext):
        pass

    # Exit a parse tree produced by kernSpineParser#ligatureTieStart.
    def exitLigatureTieStart(self, ctx:kernSpineParser.LigatureTieStartContext):
        pass


    # Enter a parse tree produced by kernSpineParser#tieContinue.
    def enterTieContinue(self, ctx:kernSpineParser.TieContinueContext):
        pass

    # Exit a parse tree produced by kernSpineParser#tieContinue.
    def exitTieContinue(self, ctx:kernSpineParser.TieContinueContext):
        pass


    # Enter a parse tree produced by kernSpineParser#ligatureTieEnd.
    def enterLigatureTieEnd(self, ctx:kernSpineParser.LigatureTieEndContext):
        pass

    # Exit a parse tree produced by kernSpineParser#ligatureTieEnd.
    def exitLigatureTieEnd(self, ctx:kernSpineParser.LigatureTieEndContext):
        pass


    # Enter a parse tree produced by kernSpineParser#slurEnd.
    def enterSlurEnd(self, ctx:kernSpineParser.SlurEndContext):
        pass

    # Exit a parse tree produced by kernSpineParser#slurEnd.
    def exitSlurEnd(self, ctx:kernSpineParser.SlurEndContext):
        pass


    # Enter a parse tree produced by kernSpineParser#barLineCrossedNoteStart.
    def enterBarLineCrossedNoteStart(self, ctx:kernSpineParser.BarLineCrossedNoteStartContext):
        pass

    # Exit a parse tree produced by kernSpineParser#barLineCrossedNoteStart.
    def exitBarLineCrossedNoteStart(self, ctx:kernSpineParser.BarLineCrossedNoteStartContext):
        pass


    # Enter a parse tree produced by kernSpineParser#barLineCrossedNoteEnd.
    def enterBarLineCrossedNoteEnd(self, ctx:kernSpineParser.BarLineCrossedNoteEndContext):
        pass

    # Exit a parse tree produced by kernSpineParser#barLineCrossedNoteEnd.
    def exitBarLineCrossedNoteEnd(self, ctx:kernSpineParser.BarLineCrossedNoteEndContext):
        pass


    # Enter a parse tree produced by kernSpineParser#stem.
    def enterStem(self, ctx:kernSpineParser.StemContext):
        pass

    # Exit a parse tree produced by kernSpineParser#stem.
    def exitStem(self, ctx:kernSpineParser.StemContext):
        pass


    # Enter a parse tree produced by kernSpineParser#beam.
    def enterBeam(self, ctx:kernSpineParser.BeamContext):
        pass

    # Exit a parse tree produced by kernSpineParser#beam.
    def exitBeam(self, ctx:kernSpineParser.BeamContext):
        pass


    # Enter a parse tree produced by kernSpineParser#mordent.
    def enterMordent(self, ctx:kernSpineParser.MordentContext):
        pass

    # Exit a parse tree produced by kernSpineParser#mordent.
    def exitMordent(self, ctx:kernSpineParser.MordentContext):
        pass


    # Enter a parse tree produced by kernSpineParser#trill.
    def enterTrill(self, ctx:kernSpineParser.TrillContext):
        pass

    # Exit a parse tree produced by kernSpineParser#trill.
    def exitTrill(self, ctx:kernSpineParser.TrillContext):
        pass


    # Enter a parse tree produced by kernSpineParser#footnote.
    def enterFootnote(self, ctx:kernSpineParser.FootnoteContext):
        pass

    # Exit a parse tree produced by kernSpineParser#footnote.
    def exitFootnote(self, ctx:kernSpineParser.FootnoteContext):
        pass



del kernSpineParser


from __future__ import annotations

from copy import deepcopy
from enum import Enum
from typing import Optional
from collections.abc import Sequence
from abc import ABC, abstractmethod

from kernpy.core import Document, SpineOperationToken, HeaderToken, Importer, TokenCategory, InstrumentToken, \
    TOKEN_SEPARATOR, DECORATION_SEPARATOR, Token, NoteRestToken, HEADERS, BEKERN_CATEGORIES, ComplexToken, Node
from kernpy.core.tokenizers import Encoding, TokenizerFactory, Tokenizer



class ExportOptions:
    """
    `ExportOptions` class.

    Store the options to export a **kern file.
    """

    def __init__(
            self,
            spine_types: [] = None,
            token_categories: [] = None,
            from_measure: int = None,
            to_measure: int = None,
            kern_type: Encoding = Encoding.normalizedKern,
            instruments: [] = None,
            show_measure_numbers: bool = False,
            spine_ids: [int] = None
    ):
        """
        Create a new ExportOptions object.

        Args:
            spine_types (Iterable): **kern, **mens, etc...
            token_categories (Iterable): TokenCategory
            from_measure (int): The measure to start exporting. When None, the exporter will start from the beginning of the file. The first measure is 1
            to_measure (int): The measure to end exporting. When None, the exporter will end at the end of the file.
            kern_type (Encoding): The type of the kern file to export.
            instruments (Iterable): The instruments to export. When None, all the instruments will be exported.
            show_measure_numbers (Bool): Show the measure numbers in the exported file.
            spine_ids (Iterable): The ids of the spines to export. When None, all the spines will be exported. Spines ids start from 0 and they are increased by 1.

        Example:
            >>> import kernpy

            Create the importer and read the file
            >>> hi = Importer()
            >>> document = hi.import_file('file.krn')
            >>> exporter = Exporter()

            Export the file with the specified options
            >>> options = ExportOptions(spine_types=['**kern'], token_categories=BEKERN_CATEGORIES)
            >>> exported_data = exporter.export_string(document, options)

            Export only the lyrics
            >>> options = ExportOptions(spine_types=['**kern'], token_categories=[TokenCategory.LYRICS])
            >>> exported_data = exporter.export_string(document, options)

            Export the comments
            >>> options = ExportOptions(spine_types=['**kern'], token_categories=[TokenCategory.LINE_COMMENTS, TokenCategory.FIELD_COMMENTS])
            >>> exported_data = exporter.export_string(document, options)

            Export using the eKern version
            >>> options = ExportOptions(spine_types=['**kern'], token_categories=BEKERN_CATEGORIES, kern_type=Encoding.eKern)
            >>> exported_data = exporter.export_string(document, options)

        """
        self.spine_types = spine_types if spine_types is not None else deepcopy(HEADERS)
        self.from_measure = from_measure
        self.to_measure = to_measure
        self.token_categories = token_categories if token_categories is not None else [c for c in TokenCategory]
        self.kern_type = kern_type
        self.instruments = instruments
        self.show_measure_numbers = show_measure_numbers
        self.spine_ids = spine_ids  # When exporting, if spine_ids=None all the spines will be exported.

    def __eq__(self, other: 'ExportOptions') -> bool:
        """
        Compare two ExportOptions objects.

        Args:
            other: The other ExportOptions object to compare.

        Returns (bool):
            True if the objects are equal, False otherwise.

        Examples:
            >>> options1 = ExportOptions(spine_types=['**kern'], token_categories=BEKERN_CATEGORIES)
            >>> options2 = ExportOptions(spine_types=['**kern'], token_categories=BEKERN_CATEGORIES)
            >>> options1 == options2
            True

            >>> options3 = ExportOptions(spine_types=['**kern', '**harm'], token_categories=BEKERN_CATEGORIES)
            >>> options1 == options3
            False
        """
        return self.spine_types == other.spine_types and \
            self.token_categories == other.token_categories and \
            self.from_measure == other.from_measure and \
            self.to_measure == other.to_measure and \
            self.kern_type == other.kern_type and \
            self.instruments == other.instruments and \
            self.show_measure_numbers == other.show_measure_numbers and \
            self.spine_ids == other.spine_ids

    def __ne__(self, other: 'ExportOptions') -> bool:
        """
        Compare two ExportOptions objects.

        Args:
            other (ExportOptions): The other ExportOptions object to compare.

        Returns (bool):
            True if the objects are not equal, False otherwise.

        Examples:
            >>> options1 = ExportOptions(spine_types=['**kern'], token_categories=BEKERN_CATEGORIES)
            >>> options2 = ExportOptions(spine_types=['**kern'], token_categories=BEKERN_CATEGORIES)
            >>> options1 != options2
            False

            >>> options3 = ExportOptions(spine_types=['**kern', '**harm'], token_categories=BEKERN_CATEGORIES)
            >>> options1 != options3
            True
        """
        return not self.__eq__(other)

    @classmethod
    def default(cls):
        return cls(
            spine_types=deepcopy(HEADERS),
            token_categories=[c for c in TokenCategory],
            from_measure=None,
            to_measure=None,
            kern_type=Encoding.normalizedKern,
            instruments=None,
            show_measure_numbers=False,
            spine_ids=None
        )


def empty_row(row):
    for col in row:
        if col != '.' and col != '' and col != '*':
            return False
    return True


class HeaderTokenGenerator:
    """
    HeaderTokenGenerator class.

    This class is used to translate the HeaderTokens to the specific encoding format.
    """
    @classmethod
    def new(cls, *, token: HeaderToken, type: Encoding):
        """
        Create a new HeaderTokenGenerator object. Only accepts stardized Humdrum **kern encodings. 

        Args:
            token (HeaderToken): The HeaderToken to be translated.
            type (Encoding): The encoding to be used.

        Examples:
            >>> header = HeaderToken('**kern', 0)
            >>> header.encoding
            '**kern'
            >>> new_header = HeaderTokenGenerator.new(token=header, type=Encoding.eKern)
            >>> new_header.encoding
            '**ekern'
        """
        new_encoding = f'**{type.prefix()}{token.encoding[2:]}'
        new_token = HeaderToken(new_encoding, token.spine_id)

        return new_token




class Exporter:
    def export_string(self, document: Document, options: ExportOptions) -> str:
        self.export_options_validator(document, options)

        rows = []

        if options.to_measure is not None and options.to_measure < len(document.measure_start_tree_stages):

            if options.to_measure < len(document.measure_start_tree_stages) - 1:
                to_stage = document.measure_start_tree_stages[
                    options.to_measure]  # take the barlines from the next coming measure
            else:
                to_stage = len(document.tree.stages) - 1  # all stages
        else:
            to_stage = len(document.tree.stages) - 1  # all stages

        if options.from_measure:
            # In case of beginning not from the first measure, we recover the spine creation and the headers
            # Traversed in reverse order to only include the active spines at the given measure...
            from_stage = document.measure_start_tree_stages[options.from_measure - 1]
            next_nodes = document.tree.stages[from_stage]
            while next_nodes and len(next_nodes) > 0 and next_nodes[0] != document.tree.root:
                row = []
                new_next_nodes = []
                non_place_holder_in_row = False
                spine_operation_row = False
                for node in next_nodes:
                    if isinstance(node.token, SpineOperationToken):
                        spine_operation_row = True
                        break

                for node in next_nodes:
                    content = ''
                    if isinstance(node.token, HeaderToken) and node.token.encoding in options.spine_types:
                        content = self.export_token(node, options)
                        non_place_holder_in_row = True
                    elif spine_operation_row:
                        # either if it is the split operator that has been cancelled, or the join one
                        if isinstance(node.token, SpineOperationToken) and (node.token.is_cancelled_at(
                                from_stage) or node.last_spine_operator_node and node.last_spine_operator_node.token.cancelled_at_stage == node.stage):
                            content = '*'
                        else:
                            content = self.export_token(node, options)
                            non_place_holder_in_row = True
                    if content:
                        row.append(content)
                    new_next_nodes.append(node.parent)
                next_nodes = new_next_nodes
                if non_place_holder_in_row:  # if the row contains just place holders due to an ommitted place holder, don't add it
                    rows.insert(0, row)

            # now, export the signatures
            node_signatures = None
            for node in document.tree.stages[from_stage]:
                node_signature_rows = []
                for signature_node in node.last_signature_nodes.nodes.values():
                    if not self.is_signature_cancelled(signature_node, node, from_stage, to_stage):
                        node_signature_rows.append(self.export_token(signature_node, options))
                if len(node_signature_rows) > 0:
                    if not node_signatures:
                        node_signatures = []  # an array for each spine
                    else:
                        if len(node_signatures[0]) != len(node_signature_rows):
                            raise Exception(f'Node signature mismatch: multiple spines with signatures at measure {len(rows)}')  # TODO better message
                    node_signatures.append(node_signature_rows)

            if node_signatures:
                for irow in range(len(node_signatures[0])):  # all spines have the same number of rows
                    row = []
                    for icol in range(len(node_signatures)):  #len(node_signatures) = number of spines
                        row.append(node_signatures[icol][irow])
                    rows.append(row)

        else:
            from_stage = 0
            rows = []

        #if not node.token.category == TokenCategory.LINE_COMMENTS and not node.token.category == TokenCategory.FIELD_COMMENTS:
        for stage in range(from_stage, to_stage + 1):  # to_stage included
            row = []
            for i_column, node in enumerate(document.tree.stages[stage]):
                self.append_row(document=document, node=node, options=options, row=row)

            nullish_tokens = {'.', '*', ''}
            if len(row) > 0 and not all(token in nullish_tokens for token in row):
                rows.append(row)

        # now, add the spine terminate row
        if options.to_measure is not None and len(rows) > 0 and rows[len(rows) - 1][
            0] != '*-':  # if the terminate is not added yet
            last_row = rows[len(rows) - 1]
            spine_count = len(last_row)
            merge_tokens_count = sum(1 for column in last_row if column == '*^')
            join_tokens_count = sum(1 for column in last_row if column == '*v')
            next_row_spine_count = spine_count + merge_tokens_count - join_tokens_count

            row = []
            for i in range(next_row_spine_count):
                row.append('*-')
            rows.append(row)

        result = ""
        for row in rows:
            if not empty_row(row):
                result += '\t'.join(row) + '\n'
        return result

    def compute_header_type(self, node) -> Optional[HeaderToken]:
        """
        Compute the header type of the node.

        Args:
            node (Node): The node to compute.

        Returns (Optional[Token]): The header type `Node`object. None if the current node is the header.

        """
        if isinstance(node.token, HeaderToken):
            header_type = node.token
        elif node.header_node:
            header_type = node.header_node.token
        else:
            header_type = None
        return header_type

    def export_token(self, node: Node, options: ExportOptions) -> str:
        token = node.token
        if isinstance(token, HeaderToken):
            new_token = HeaderTokenGenerator.new(token=token, type=options.kern_type)
        else:
            new_token = token

        last_clef_node = node.last_signature_nodes.nodes.get('ClefToken', None)
        if last_clef_node is not None:
            last_clef = last_clef_node.token
        else:
            last_clef = None  # Any clef appears at this point of the score (e.g., metadata rows)

        return (TokenizerFactory
                .create(options.kern_type.value, token_categories=options.token_categories, last_clef_reference=last_clef)
                .tokenize(new_token))

    def append_row(self, document: Document, node, options: ExportOptions, row: list) -> bool:
        """
        Append a row to the row list if the node accomplishes the requirements.
        Args:
            document (Document): The document with the spines.
            node (Node): The node to append.
            options (ExportOptions): The export options to filter the token.
            row (list): The row to append.

        Returns (bool): True if the row was appended. False if the row was not appended.
        """
        header_type = self.compute_header_type(node)

        if not (header_type is not None
                and header_type.encoding in options.spine_types
                and (options.spine_ids is None or header_type.spine_id in options.spine_ids)
        ):
            return False  # All the spine must be filtered out

        if not (not node.token.hidden
                and (isinstance(node.token, ComplexToken) or node.token.category in options.token_categories)
                # If None, all the spines will be exported. TODO: put all the spines as spine_ids = None
        ):
            row.append(self._retrieve_empty_token(node))
            return True  # The spine must be kept, but this specific token does not achieve the requirements

        # Normal case
        exported_token = self.export_token(node, options)
        exported_token = exported_token if len(exported_token) > 0 else self._retrieve_empty_token(node) # just in the unexpected case, the tokenizer returns an empty string...
        row.append(exported_token)
        return True


    @classmethod
    def _is_token_in_a_signature_row(cls, node: Node) -> bool:
        return bool(TokenCategory.is_child(
            child=node.token.category,
            parent=TokenCategory.SIGNATURES,
        ))

    @classmethod
    def _retrieve_empty_token(cls, node: Optional[Node]) -> str:
        if node is None or node.token is None:
            return ''
        return '.' if not cls._is_token_in_a_signature_row(node) else '*'

    def get_spine_types(self, document: Document, spine_types: list = None):
        """
        Get the spine types from the document.

        Args:
            document (Document): The document with the spines.
            spine_types (list): The spine types to export. If None, all the spine types will be exported.

        Returns: A list with the spine types.

        Examples:
            >>> exporter = Exporter()
            >>> exporter.get_spine_types(document)
            ['**kern', '**kern', '**kern', '**kern', '**root', '**harm']
            >>> exporter.get_spine_types(document, None)
            ['**kern', '**kern', '**kern', '**kern', '**root', '**harm']
            >>> exporter.get_spine_types(document, ['**kern'])
            ['**kern', '**kern', '**kern', '**kern']
            >>> exporter.get_spine_types(document, ['**kern', '**root'])
            ['**kern', '**kern', '**kern', '**kern', '**root']
            >>> exporter.get_spine_types(document, ['**kern', '**root', '**harm'])
            ['**kern', '**kern', '**kern', '**kern', '**root', '**harm']
            >>> exporter.get_spine_types(document, [])
            []
        """
        if spine_types is not None and len(spine_types) == 0:
            return []

        options = ExportOptions(spine_types=spine_types, token_categories=[TokenCategory.HEADER])
        content = self.export_string(document, options)

        # Remove all after the first line: **kern, **mens, etc... are always in the first row
        lines = content.split('\n')
        first_line = lines[0:1]
        tokens = first_line[0].split('\t')

        return tokens if tokens not in [[], ['']] else []


    @classmethod
    def export_options_validator(cls, document: Document, options: ExportOptions) -> None:
        """
        Validate the export options. Raise an exception if the options are invalid.

        Args:
            document: `Document` - The document to export.
            options: `ExportOptions` - The options to export the document.

        Returns: None

        Example:
            >>> export_options_validator(document, options)
            ValueError: option from_measure must be >=0 but -1 was found.
            >>> export_options_validator(document, options2)
            None
        """
        if options.from_measure is not None and options.from_measure < 0:
            raise ValueError(f'option from_measure must be >=0 but {options.from_measure} was found. ')
        if options.to_measure is not None and options.to_measure > len(document.measure_start_tree_stages):
            # "TODO: DAVID, check options.to_measure bounds. len(document.measure_start_tree_stages) or len(document.measure_start_tree_stages) - 1"
            raise ValueError(
                f'option to_measure must be <= {len(document.measure_start_tree_stages)} but {options.to_measure} was found. ')
        if options.to_measure is not None and options.from_measure is not None and options.to_measure < options.from_measure:
            raise ValueError(
                f'option to_measure must be >= from_measure but {options.to_measure} < {options.from_measure} was found. ')

    def is_signature_cancelled(self, signature_node, node, from_stage, to_stage) -> bool:
        if node.token.__class__ == signature_node.token.__class__:
            return True
        elif isinstance(node.token, NoteRestToken):
            return False
        elif from_stage < to_stage:
            for child in node.children:
                if self.is_signature_cancelled(signature_node, child, from_stage + 1, to_stage):
                    return True
            return False


def get_kern_from_ekern(ekern_content: str) -> str:
    """
    Read the content of a **ekern file and return the **kern content.

    Args:
        ekern_content: The content of the **ekern file.
    Returns:
        The content of the **kern file.

    Example:
        ```python
        # Read **ekern file
        ekern_file = 'path/to/file.ekrn'
        with open(ekern_file, 'r') as file:
            ekern_content = file.read()

        # Get **kern content
        kern_content = get_kern_from_ekern(ekern_content)
        with open('path/to/file.krn', 'w') as file:
            file.write(kern_content)

        ```
    """
    content = ekern_content.replace("**ekern", "**kern")  # TODO Constante según las cabeceras
    content = content.replace(TOKEN_SEPARATOR, "")
    content = content.replace(DECORATION_SEPARATOR, "")

    return content


def ekern_to_krn(
        input_file: str,
        output_file: str
) -> None:
    """
    Convert one .ekrn file to .krn file.

    Args:
        input_file (str): Filepath to the input **ekern
        output_file (str): Filepath to the output **kern
    Returns:
        None

    Example:
        # Convert .ekrn to .krn
        >>> ekern_to_krn('path/to/file.ekrn', 'path/to/file.krn')

        # Convert a list of .ekrn files to .krn files
        ```python
        ekrn_files = your_modue.get_files()

        # Use the wrapper to avoid stopping the process if an error occurs
        def ekern_to_krn_wrapper(ekern_file, kern_file):
            try:
                ekern_to_krn(ekrn_files, output_folder)
            except Exception as e:
                print(f'Error:{e}')

        # Convert all the files
        for ekern_file in ekrn_files:
            output_file = ekern_file.replace('.ekrn', '.krn')
            ekern_to_krn_wrapper(ekern_file, output_file)
        ```
    """
    with open(input_file, 'r') as file:
        content = file.read()

    kern_content = get_kern_from_ekern(content)

    with open(output_file, 'w') as file:
        file.write(kern_content)


def kern_to_ekern(
        input_file: str,
        output_file: str
) -> None:
    """
    Convert one .krn file to .ekrn file

    Args:
        input_file (str): Filepath to the input **kern
        output_file (str): Filepath to the output **ekern

    Returns:
        None

    Example:
        # Convert .krn to .ekrn
        >>> kern_to_ekern('path/to/file.krn', 'path/to/file.ekrn')

        # Convert a list of .krn files to .ekrn files
        ```python
        krn_files = your_module.get_files()

        # Use the wrapper to avoid stopping the process if an error occurs
        def kern_to_ekern_wrapper(krn_file, ekern_file):
            try:
                kern_to_ekern(krn_file, ekern_file)
            except Exception as e:
                print(f'Error:{e}')

        # Convert all the files
        for krn_file in krn_files:
            output_file = krn_file.replace('.krn', '.ekrn')
            kern_to_ekern_wrapper(krn_file, output_file)
        ```

    """
    importer = Importer()
    document = importer.import_file(input_file)

    if len(importer.errors):
        raise Exception(f'ERROR: {input_file} has errors {importer.get_error_messages()}')

    export_options = ExportOptions(spine_types=['**kern'], token_categories=BEKERN_CATEGORIES,
                                   kern_type=Encoding.eKern)
    exporter = Exporter()
    exported_ekern = exporter.export_string(document, export_options)

    with open(output_file, 'w') as file:
        file.write(exported_ekern)



import os
import unittest

import kernpy as kp


class ExporterTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Read the basic score document once before all tests
        cls.doc_organ_4_voices, _ = kp.read('resource_dir/legacy/chor048.krn')
        cls.doc_piano, _ = kp.read('resource_dir/mozart/concerto-piano-12-allegro.krn')

    def _run_HeaderTokenGenerator_new_generalized(self, input_token, export_type, expected):
        """Private helper method to run individual test cases."""
        token = kp.HeaderToken(input_token, 0)
        result = kp.HeaderTokenGenerator.new(token=token, type=export_type).encoding
        self.assertEqual(
            result, expected,
            f"Failed for input: {input_token}, export_type: {export_type}, expected: {expected}, got: {result}"
        )

    def exported_filtering_by_category(self, doc, include, exclude, expected_path):
        # Arrange
        with open(expected_path, 'r') as f:
            expected_content = f.read()

        # Act
        real_content = kp.dumps(doc, include=include, exclude=exclude)

        # Assert
        self.assertEqual(expected_content, real_content)


    def test_get_spine_types_1(self):
        spine_types = kp.spine_types(self.doc_organ_4_voices)
        self.assertEqual(['**kern', '**kern', '**kern', '**kern', '**root', '**harm'], spine_types)

    def test_get_spine_types_2(self):
        spine_types = kp.get_spine_types(self.doc_organ_4_voices, spine_types=None)
        self.assertEqual(['**kern', '**kern', '**kern', '**kern', '**root', '**harm'], spine_types)

    def test_get_spine_types_3(self):
        spine_types = kp.get_spine_types(self.doc_organ_4_voices, spine_types=['**kern'])
        self.assertEqual(['**kern', '**kern', '**kern', '**kern'], spine_types)

    def test_get_spine_types_4(self):
        spine_types = kp.get_spine_types(self.doc_organ_4_voices, spine_types=['**root'])
        self.assertEqual(['**root'], spine_types)

    def test_get_spine_types_5(self):
        spine_types = kp.get_spine_types(self.doc_organ_4_voices, spine_types=['**harm'])
        self.assertEqual(['**harm'], spine_types)

    def test_get_spine_types_6(self):
        spine_types = kp.get_spine_types(self.doc_organ_4_voices, spine_types=['**not-exists'])
        self.assertEqual([], spine_types)

    def test_get_spine_types_7(self):
        spine_types = kp.get_spine_types(self.doc_organ_4_voices, spine_types=[])
        self.assertEqual([], spine_types)

    def test_exporter_spine_id_left_hand(self):
        with open('resource_dir/mozart/concerto-piano-12-allegro-left-hand.krn', 'r') as f:
            expected_content = f.read()
        real_content = kp.dumps(self.doc_piano, spine_ids=[0])
        print(f"Real content:\n{real_content}\nExpected content:\n{expected_content}")
        self.assertEqual(expected_content, real_content)

    def test_exporter_spine_id_right_hand(self):
        with open('resource_dir/mozart/concerto-piano-12-allegro-right-hand.krn', 'r') as f:
            expected_content = f.read()
        real_content = kp.dumps(self.doc_piano, spine_ids=[1])
        self.assertEqual(expected_content, real_content)

    def test_exporter_spines_id_mix_kern_dyn_spines(self):
        with open('resource_dir/mozart/concerto-piano-12-allegro-right-hand-and-dyn.krn', 'r') as f:
            expected_content = f.read()
        real_content = kp.dumps(self.doc_piano,
                                 spine_ids=[0, 2])  # TODO: Solve export error: error in line 15 of the exported file. No tiene nada que ver con la funcionalidad de exportar por spines. Sino con exportar todo en general.
        self.assertEqual(expected_content, real_content)

    def test_basic_kern_to_ekern(self):
        input_path = 'resource_dir/legacy/kern2ekern.krn'
        expected_path = 'resource_dir/legacy/kern2ekern.ekrn'

        with open(expected_path, 'r') as f:
            expected_content = f.read()

        doc, _ = kp.load(input_path)
        real_content = kp.dumps(doc, encoding=kp.Encoding.eKern, include=kp.BEKERN_CATEGORIES)

        self.assertEqual(expected_content, real_content,
                         f"File content mismatch: \nExpected:\n{expected_content}\n{40 * '='}\nReal\n{real_content}")

    def test_check_categories_are_exported(self):
        data = self.doc_organ_4_voices.frequencies([kp.TokenCategory.CORE])
        for k, v in data.items():
            print(f"{k}: {v}")

    def test_should_export_with_all(self):
        self.exported_filtering_by_category(
            doc=self.doc_piano,
            include={t for t in kp.TokenCategory},
            exclude=None,
            expected_path='resource_dir/categories/concerto-piano-12-allegro_with_all.krn'
        )

    def test_should_export_without_barlines(self):
        self.exported_filtering_by_category(
            doc=self.doc_piano,
            include={t for t in kp.TokenCategory},
            exclude={kp.TokenCategory.BARLINES},
            expected_path='resource_dir/categories/concerto-piano-12-allegro_without_barlines.krn'
        )

    def test_should_export_without_only_signatures(self):
        self.exported_filtering_by_category(
            doc=self.doc_piano,
            include={kp.TokenCategory.SIGNATURES},
            exclude=None,
            expected_path='resource_dir/categories/concerto-piano-12-allegro_without_only_signatures.krn'
        )

    def test_should_export_without_harmony(self):
        self.exported_filtering_by_category(
            doc=self.doc_organ_4_voices,
            include={t for t in kp.TokenCategory},
            exclude={kp.TokenCategory.HARMONY},
            expected_path='resource_dir/categories/concerto-piano-12-allegro_without_harmony.krn'
        )

    def test_only_export_kern_and_harm_spines(self):
        with open('resource_dir/spines/concerto-piano-12-allegro_only_kern_and_harm.krn', 'r') as f:
            expected_content = f.read()

        real_content = kp.dumps(self.doc_organ_4_voices, spine_types=['**kern', '**harm'])

        self.assertEqual(expected_content, real_content,
                         f"File content mismatch: \nExpected:\n{expected_content}\n{40 * '='}\nReal\n{real_content}")

    def test_exporter_HeaderTokenGenerator_new(self):
        self._run_HeaderTokenGenerator_new_generalized("**kern", kp.Encoding.eKern, "**ekern")
        self._run_HeaderTokenGenerator_new_generalized("**kern", kp.Encoding.normalizedKern, "**kern")
        self._run_HeaderTokenGenerator_new_generalized("**kern", kp.Encoding.bKern, "**bkern")
        self._run_HeaderTokenGenerator_new_generalized("**kern", kp.Encoding.bEkern, "**bekern")







"""
Public API for KernPy.

The main functions for handling the input and output of **kern files are provided here.
"""
from __future__ import annotations

from pathlib import Path
from typing import List, Optional, Any, Union, Tuple, Sequence

from kernpy import Encoding
from kernpy.core import (
    Document, Importer, Exporter, ExportOptions, GraphvizExporter,
    generic,
    TokenCategoryHierarchyMapper,
    TokenCategory,
)


def load(fp: Union[str, Path], *, raise_on_errors: Optional[bool] = False, **kwargs) -> (Document, List[str]):
    """
    Load a Document object from a Humdrum **kern file-like object.

    Args:
        fp (Union[str, Path]): A path-like object representing a **kern file.
        raise_on_errors (Optional[bool], optional): If True, raise an exception if any grammar error is detected\
            during parsing.

    Returns ((Document, List[str])): A tuple containing the Document object and a list of messages representing \
        grammar errors detected during parsing. If the list is empty,\
        the parsing did not detect any errors.

    Raises:
        ValueError: If the Humdrum **kern representation could not be parsed.

    Examples:
        >>> import kernpy as kp
        >>> document, errors = kp.load('BWV565.krn')
        >>> if len(errors) > 0:
        >>>     print(f"Grammar didn't recognize the following errors: {errors}")
        ['Error: Invalid **kern spine: 1', 'Error: Invalid **kern spine: 2']
        >>>     # Anyway, we can use the Document
        >>>     print(document)
        >>> else:
        >>>     print(document)
        <kernpy.core.document.Document object at 0x7f8b3b7b3d90>
    """
    return generic.Generic.read(
        path=fp,
        strict=raise_on_errors,
    )


def loads(s, *, raise_on_errors: Optional[bool] = False, **kwargs) -> (Document, List[str]):
    """
    Load a Document object from a string encoded in Humdrum **kern.

    Args:
        s (str): A string containing a **kern file.
        raise_on_errors (Optional[bool], optional): If True, raise an exception if any grammar error is detected\
            during parsing.

    Returns ((Document, List[str])): A tuple containing the Document object and a list of messages representing \
        grammar errors detected during parsing. If the list is empty,\
        the parsing did not detect any errors.

    Raises:
        ValueError: If the Humdrum **kern representation could not be parsed.

    Examples:
        >>> import kernpy as kp
        >>> document, errors = kp.loads('**kern\n*clefG2\n=1\n4c\n4d\n4e\n4f\n')
        >>> if len(errors) > 0:
        >>>     print(f"Grammar didn't recognize the following errors: {errors}")
        ['Error: Invalid **kern spine: 1']
        >>>     # Anyway, we can use the Document
        >>>     print(document)
        >>> else:
        >>>     print(document)
        <kernpy.core.document.Document object at 0x7f8b3b7b3d90>
    """
    return generic.Generic.create(
        content=s,
        strict=raise_on_errors,
    )


def dump(document: Document, fp: Union[str, Path], *,
         spine_types: [str] = None,
         include: [TokenCategory] = None,
         exclude: [TokenCategory] = None,
         from_measure: int = None,
         to_measure: int = None,
         encoding: Encoding = None,
         instruments: [str] = None,
         show_measure_numbers: bool = None,
         spine_ids: [int] = None
         ) -> None:
    """

    Args:
        document (Document): The Document object to write to the file.
        fp (Union[str, Path]): The file path to write the Document object.
        spine_types (Iterable): **kern, **mens, etc...
        include (Iterable): The token categories to include in the exported file. When None, all the token categories will be exported.
        exclude (Iterable): The token categories to exclude from the exported file. When None, no token categories will be excluded.
        from_measure (int): The measure to start exporting. When None, the exporter will start from the beginning of the file. The first measure is 1
        to_measure (int): The measure to end exporting. When None, the exporter will end at the end of the file.
        encoding (Encoding): The type of the **kern file to export.
        instruments (Iterable): The instruments to export. If None, all the instruments will be exported.
        show_measure_numbers (Bool): Show the measure numbers in the exported file.
        spine_ids (Iterable): The ids of the spines to export. When None, all the spines will be exported. \
            Spines ids start from 0, and they are increased by 1 for each spine to the right.


    Returns (None): None

    Raises:
        ValueError: If the document could not be exported.

    Examples:
        >>> import kernpy as kp
        >>> document, errors = kp.load('BWV565.krn')
        >>> kp.dump(document, 'BWV565_normalized.krn')
        None
        >>> # File 'BWV565_normalized.krn' will be created with the normalized **kern representation.
    """
    # Create an ExportOptions instance with only user-modified arguments
    options = generic.Generic.parse_options_to_ExportOptions(
        spine_types=spine_types,
        include=include,
        exclude=exclude,
        from_measure=from_measure,
        to_measure=to_measure,
        kern_type=encoding,
        instruments=instruments,
        show_measure_numbers=show_measure_numbers,
        spine_ids=spine_ids
    )

    return generic.Generic.store(
        document=document,
        path=fp,
        options=options
    )


def dumps(document: Document, *,
          spine_types: [str] = None,
          include: [TokenCategory] = None,
          exclude: [TokenCategory] = None,
          from_measure: int = None,
          to_measure: int = None,
          encoding: Encoding = None,
          instruments: [str] = None,
          show_measure_numbers: bool = None,
          spine_ids: [int] = None
          ) -> str:
    """

    Args:
        document (Document): The Document object to write to the file.
        fp (Union[str, Path]): The file path to write the Document object.
        spine_types (Iterable): **kern, **mens, etc...
        include (Iterable): The token categories to include in the exported file. When None, all the token categories will be exported.
        exclude (Iterable): The token categories to exclude from the exported file. When None, no token categories will be excluded.
        from_measure (int): The measure to start exporting. When None, the exporter will start from the beginning of the file. The first measure is 1
        to_measure (int): The measure to end exporting. When None, the exporter will end at the end of the file.
        encoding (Encoding): The type of the **kern file to export.
        instruments (Iterable): The instruments to export. If None, all the instruments will be exported.
        show_measure_numbers (Bool): Show the measure numbers in the exported file.
        spine_ids (Iterable): The ids of the spines to export. When None, all the spines will be exported. \
            Spines ids start from 0, and they are increased by 1 for each spine to the right.


    Returns (None): None

    Raises:
        ValueError: If the document could not be exported.

    Examples:
        >>> import kernpy as kp
        >>> document, errors = kp.load('score.krn')
        >>> kp.dumps(document)
        '**kern\n*clefG2\n=1\n4c\n4d\n4e\n4f\n*-'
    """
    # Create an ExportOptions instance with only user-modified arguments
    options = generic.Generic.parse_options_to_ExportOptions(
        spine_types=spine_types,
        include=include,
        exclude=exclude,
        from_measure=from_measure,
        to_measure=to_measure,
        kern_type=encoding,
        instruments=instruments,
        show_measure_numbers=show_measure_numbers,
        spine_ids=spine_ids
    )

    return generic.Generic.export(
        document=document,
        options=options
    )


def graph(document: Document, fp: Optional[Union[str, Path]]) -> None:
    """
    Create a graph representation of a Document object using Graphviz. Save the graph as a .dot file or indicate the\
     output file path or stream. If the output file path is None, the function will return the graphviz content as a\
        string to the standard output.

    Use the Graphviz software to convert the .dot file to an image.


    Args:
        document (Document): The Document object to export as a graphviz file.
        fp (Optional[Union[str, Path]]): The file path to write the graphviz file. If None, the function will return the\
            graphviz content as a string to the standard output.

    Returns (None): None

    Examples:
        >>> import kernpy as kp
        >>> document, errors = kp.load('score.krn')
        >>> kp.graph(document, 'score.dot')
        None
        >>> # File 'score.dot' will be created with the graphviz representation of the Document object.
        >>> kp.graph(document, None)
        'digraph G { ... }'
    """
    return generic.Generic.store_graph(
        document=document,
        path=fp
    )


def concat(
        contents: List[str],
        *,
        separator: Optional[str] = '\n',
) -> Tuple[Document, List[Tuple[int, int]]]:
    """
    Concatenate multiple **kern fragments into a single Document object. \
     All the fragments should be presented in order. Each fragment does not need to be a complete **kern file. \

    Warnings:
        Processing a large number of files in a row may take some time.
         This method performs as many `kp.read` operations as there are fragments to concatenate.
    Args:
        contents (Sequence[str]): List of **kern strings
        separator (Optional[str]): Separator string to separate the **kern fragments. Default is '\n' (newline).

    Returns (Tuple[Document, List[Tuple[int, int]]]): Document object and \
      and a List of Pairs (Tuple[int, int]) representing the measure fragment indexes of the concatenated document.

    Examples:
        >>> import kernpy as kp
        >>> contents = ['**kern\n4e\n4f\n4g\n*-\n', '4a\n4b\n4c\n*-\n=\n', '4d\n4e\n4f\n*-\n']
        >>> document, indexes = kp.concat(contents)
        >>> indexes
        [(0, 3), (3, 6), (6, 9)]
        >>> document, indexes = kp.concat(contents, separator='\n')
        >>> indexes
        [(0, 3), (3, 6), (6, 9)]
        >>> document, indexes = kp.concat(contents, separator='')
        >>> indexes
        [(0, 3), (3, 6), (6, 9)]
        >>> for start, end in indexes:
        >>>     print(kp.dumps(document, from_measure=start, to_measure=end)))
    """
    return generic.Generic.concat(
        contents=contents,
        separator=separator,
    )


def merge(
        contents: List[str],
        *,
        raise_on_errors: Optional[bool] = False,
) -> Tuple[Document, List[Tuple[int, int]]]:
    """
    Merge multiple **kern fragments into a single **kern string. \
     All the fragments should be presented in order. Each fragment does not need to be a complete **kern file. \

    Warnings:
        Processing a large number of files in a row may take some time.
         This method performs as many `kp.read` operations as there are fragments to concatenate.
    Args:
        contents (Sequence[str]): List of **kern strings
        raise_on_errors (Optional[bool], optional): If True, raise an exception if any grammar error is detected\
            during parsing.

    Returns (Tuple[Document, List[Tuple[int, int]]]): Document object and \
      and a List of Pairs (Tuple[int, int]) representing the measure fragment indexes of the concatenated document.

    Examples:
        >>> import kernpy as kp
        >>> contents = ['**kern\n4e\n4f\n4g\n*-\n*-', '**kern\n4a\n4b\n4c\n*-\n=\n*-', '**kern\n4d\n4e\n4f\n*-\n*-']
        >>> document, indexes = kp.concat(contents)
        >>> indexes
        [(0, 3), (3, 6), (6, 9)]
        >>> document, indexes = kp.concat(contents, separator='\n')
        >>> indexes
        [(0, 3), (3, 6), (6, 9)]
        >>> document, indexes = kp.concat(contents, separator='')
        >>> indexes
        [(0, 3), (3, 6), (6, 9)]
        >>> for start, end in indexes:
        >>>     print(kp.dumps(document, from_measure=start, to_measure=end)))
    """
    return generic.Generic.merge(
        contents=contents,
        strict=raise_on_errors
    )


def spine_types(
        document: Document,
        headers: Optional[Sequence[str]] = None
) -> List[str]:
    """
    Get the spines of a Document object.

    Args:
        document (Document): Document object to get spines from
        headers (Optional[Sequence[str]]): List of spine types to get. If None, all spines are returned. Using a \
         header will return all the spines of that type.

    Returns (List[str]): List of spines

    Examples:
        >>> import kernpy as kp
        >>> document, _ = kp.read('path/to/file.krn')
        >>> kp.spine_types(document)
        ['**kern', '**kern', '**kern', '**kern', '**root', '**harm']
        >>> kp.spine_types(document, None)
        ['**kern', '**kern', '**kern', '**kern', '**root', '**harm']
        >>> kp.spine_types(document, headers=None)
        ['**kern', '**kern', '**kern', '**kern', '**root', '**harm']
        >>> kp.spine_types(document, headers=['**kern'])
        ['**kern', '**kern', '**kern', '**kern']
        >>> kp.spine_types(document, headers=['**kern', '**root'])
        ['**kern', '**kern', '**kern', '**kern', '**root']
        >>> kp.spine_types(document, headers=['**kern', '**root', '**harm'])
        ['**kern', '**kern', '**kern', '**kern', '**root', '**harm']
        >>> kp.spine_types(document, headers=[])
        []
    """
    return generic.Generic.get_spine_types(
        document=document,
        spine_types=headers
    )



def is_monophonic(
        document: Document,
) -> bool:
    """
    Check if a Document object is monophonic. Checks if the Document object has only one **kern spine, \
    no chord tokens, and at least one note or rest token. \

    Args:
        document (Document): Document object to check


    Returns (bool): True if the Document object is monophonic, False otherwise.

    Examples:
        >>> import kernpy as kp
        >>> document_a, _ = kp.load('path/to/monophonic-file.krn')
        >>> kp.is_monophonic(document_a)
        True
        >>> document_b, _ = kp.load('path/to/polyphonic-file.krn')
        >>> kp.is_monophonic(document_b)
        False
    """
    number_of_kern_spines = len(spine_types(document, headers=['**kern']))
    number_of_chord_tokens = len(document.get_all_tokens(filter_by_categories=[TokenCategory.CHORD]))
    there_is_any_note_rest_token = len(document.get_all_tokens(filter_by_categories=[TokenCategory.NOTE_REST])) > 0

    return (number_of_kern_spines == 1
            and number_of_chord_tokens == 0
            and there_is_any_note_rest_token)




import os
import unittest
import logging
import sys
from pathlib import Path
from tempfile import TemporaryDirectory
import json

import kernpy as kp


class DocumentTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Read the basic score document once before all tests
        cls.doc_organ_4_voices, _ = kp.read('resource_dir/legacy/chor048.krn')
        cls.doc_piano, _ = kp.read('resource_dir/mozart/concerto-piano-12-allegro.krn')

    def all_tokens_have_the_correct_category(self, doc: kp.Document, category: kp.TokenCategory):
        tokens = doc.get_all_tokens(filter_by_categories=[category])
        for token in tokens:
            self.assertEqual(category, token.category)

    @unittest.skip
    def test_document_self_concat(self):
        # Arrange
        doc_a, _ = kp.read('resource_dir/legacy/base_tuplet.krn')
        doc_b, _ = kp.read('resource_dir/legacy/base_tuplet_longer.krn')

        doc_concat = kp.Document.to_concat(doc_a, doc_b)
        kp.store_graph(doc_concat, '/tmp/graph_concat.dot')
        kp.store_graph(doc_a, '/tmp/graph_a.dot')
        kp.store_graph(doc_b, '/tmp/graph_b.dot')

        # compare
        ...

    @unittest.skip("TODO: Not implemented yet")
    def test_document_append_spines(self):
        doc, _ = kp.read('resource_dir/legacy/base_tuplet.krn')
        doc.append_spines(spines=['4e\t4f\t4g\t4a\n4b\t4c\t4d\t4e\n*_\t*_\t*_\t*_\n'])

        pass

    @unittest.skip("TODO: Not implemented yet")
    def test_document_get_voices(self):
        doc, _ = kp.read('resource_dir/legacy/chor048.krn')

        voices = doc.get_voices()
        self.assertEqual(['!sax', '!piano', '!bass'], voices)

        voices = doc.get_voices(clean=True)
        self.assertEqual(['sax', 'piano', 'bass'], voices)

    def test_document_get_header_nodes(self):
        input_kern_file = 'resource_dir/mozart/divertimento-quartet.krn'
        doc, err = kp.read(input_kern_file)
        headers_nodes = doc.get_header_nodes()

        self.assertEqual(8, len(headers_nodes))
        for node in headers_nodes:
            self.assertIsInstance(node, kp.core.HeaderToken)
        self.assertListEqual(
            ['**kern', '**dynam', '**kern', '**dynam', '**kern', '**dynam', '**kern', '**dynam'],
            [t.encoding for t in headers_nodes])

    def test_document_get_all_spines_ids(self):
        input_kern_file = 'resource_dir/mozart/divertimento-quartet.krn'

        doc, err = kp.read(input_kern_file)
        spines_ids = doc.get_spine_ids()
        self.assertListEqual(
            [0, 1, 2, 3, 4, 5, 6, 7],
            spines_ids)

    def test_document_maximum_recursion_depth_exceeded_in_tree_traversal_dfs(self):
        doc, err = kp.read('resource_dir/samples/score_with_dividing_two_spines.krn')
        tokens = doc.get_all_tokens(filter_by_categories=[kp.TokenCategory.CORE])

        # optimize this dfs implementation for not exceeding maximum recursion depth
        # use memoization to speed up the process
        # use queues
        self.assertTrue(len(tokens) > 0)

    def test_frequencies(self):
        with open('resource_dir/metadata/frequency.json', 'r') as f:
            expected_frequencies = json.load(f)

        doc, err = kp.read('resource_dir/legacy/chor001.krn')
        real_frequencies = doc.frequencies()

        self.assertDictEqual(expected_frequencies, real_frequencies)

    def test_document_split(self):
        doc, err = kp.read('resource_dir/legacy/chor001.krn')
        docs = doc.split()

        for i, new_doc in enumerate(docs):
            kp.store_graph(new_doc, f'/tmp/test_{i}.dot')
            kp.store(new_doc, f'/tmp/test_{i}.krn', kp.ExportOptions())
        self.assertEqual(4, len(docs))

    def test_has_categories_when_export(self):
        data: dict = self.doc_piano.frequencies({t for t in kp.TokenCategory} - {kp.TokenCategory.BARLINES})
        for k, v in data.items():
            print(f"{k}: {v}")

    def test_should_not_detect_barlines_tokens_with_non_barlines_category_0(self):
        frequencies = self.doc_organ_4_voices.frequencies()
        self.assertEqual(kp.TokenCategory.BARLINES.name, frequencies['=']['category'])

    def test_should_not_detect_barlines_tokens_with_non_barlines_category_1(self):
        frequencies = self.doc_organ_4_voices.frequencies()
        with self.assertRaises(KeyError):
            self.assertEqual(kp.TokenCategory.BARLINES.name, frequencies['=1-']['category'])
            self.assertEqual(kp.TokenCategory.BARLINES.name, frequencies['=2']['category'])
            self.assertEqual(kp.TokenCategory.BARLINES.name, frequencies['=3']['category'])
            self.assertEqual(kp.TokenCategory.BARLINES.name, frequencies['=4']['category'])
            ...

    def test_should_not_detect_barlines_tokens_with_non_barlines_category_2(self):
        frequencies = self.doc_organ_4_voices.frequencies()
        self.assertEqual(kp.TokenCategory.BARLINES.name, frequencies['====']['category'])

    def test_should_not_detect_barlines_tokens_with_non_barlines_category_3(self):
        frequencies = self.doc_piano.frequencies()
        self.assertEqual(kp.TokenCategory.BARLINES.name, frequencies['==:|!']['category'])

    def test_should_not_detect_barlines_tokens_with_non_barlines_category_4(self):
        frequencies = self.doc_piano.frequencies()
        with self.assertRaises(KeyError):
            self.assertEqual(kp.TokenCategory.BARLINES.name, frequencies['=94:|!|:']['category'])

    def test_all_tokens_have_correct_category_get_all_tokens_core(self):
        self.all_tokens_have_the_correct_category(self.doc_organ_4_voices, kp.TokenCategory.CORE)

    def test_all_tokens_have_correct_category_get_all_tokens_barlines(self):
        self.all_tokens_have_the_correct_category(self.doc_organ_4_voices, kp.TokenCategory.BARLINES)

    def test_all_tokens_have_correct_category_get_all_tokens_dynamics(self):
        self.all_tokens_have_the_correct_category(self.doc_organ_4_voices, kp.TokenCategory.DYNAMICS)

    def test_all_tokens_have_correct_category_get_all_tokens_signatures(self):
        self.all_tokens_have_the_correct_category(self.doc_organ_4_voices, kp.TokenCategory.SIGNATURES)

    def test_all_tokens_have_correct_category_get_all_tokens_fingering(self):
        self.all_tokens_have_the_correct_category(self.doc_organ_4_voices, kp.TokenCategory.FINGERING)

    def test_all_tokens_have_correct_category_get_all_tokens_lyrics(self):
        self.all_tokens_have_the_correct_category(self.doc_organ_4_voices, kp.TokenCategory.LYRICS)

    def test_all_tokens_have_correct_category_get_all_tokens_lines_comments(self):
        self.all_tokens_have_the_correct_category(self.doc_organ_4_voices, kp.TokenCategory.LINE_COMMENTS)

    def test_all_tokens_have_correct_category_get_all_tokens_field_comments(self):
        self.all_tokens_have_the_correct_category(self.doc_organ_4_voices, kp.TokenCategory.FIELD_COMMENTS)

    def test_match_same_document_reference(self):
        self.assertTrue(self.doc_organ_4_voices.match(self.doc_organ_4_voices, self.doc_organ_4_voices))

    def test_match_different_document_reference(self):
        self.assertFalse(self.doc_organ_4_voices.match(self.doc_organ_4_voices, self.doc_piano))

    def test_match_different_references_same_content(self):
        doc_1, _ = kp.loads('**kern\n=1\n=2\n=3\n=4\n*-\n')
        doc_2, _ = kp.loads('**kern\n=1\n=2\n=3\n=4\n*-\n')

        self.assertTrue(kp.Document.match(doc_1, doc_2))

    def test_match_different_references_different_content(self):
        doc_1, _ = kp.loads('**kern\n=1\n=2\n=3\n=4\n*-\n')
        doc_2, _ = kp.loads('**kern\t**kern\n=1\t=1\n=2\t=2\n=3\t=3\n=4\t=4\n*-\t*-\n')

        self.assertFalse(kp.Document.match(doc_1, doc_2))

    def test_match_same_kern_different_others_default_option(self):
        doc_1, _ = kp.loads('**kern\n=1\n=2\n=3\n=4\n*-\n')
        doc_2, _ = kp.loads('**kern\t**dynam\n=1\tpp\n=2\tpp\n=3\tpp\n=4\tpp\n*-\t*-\n')

        self.assertFalse(kp.Document.match(doc_1, doc_2))

    def test_match_same_kern_different_others_default_only_core(self):
        doc_1, _ = kp.loads('**kern\n=1\n=2\n=3\n=4\n*-\n')
        doc_2, _ = kp.loads('**kern\t**dynam\n=1\tpp\n=2\tpp\n=3\tpp\n=4\tpp\n*-\t*-\n')

        self.assertFalse(kp.Document.match(doc_1, doc_2, check_core_spines_only=False))

    def test_match_same_kern_different_others_default_check_all_core(self):
        doc_1, _ = kp.loads('**kern\n=1\n=2\n=3\n=4\n*-\n')
        doc_2, _ = kp.loads('**kern\t**dynam\n=1\tpp\n=2\tpp\n=3\tpp\n=4\tpp\n*-\t*-\n')

        self.assertTrue(kp.Document.match(doc_1, doc_2, check_core_spines_only=True))

    def test_add_document(self):
        doc_1, _ = kp.loads('**kern\n=1\n=2\n=3\n=4\n*-\n')
        doc_2, _ = kp.loads('**kern\n=5\n=6\n=7\n=8\n*-\n')

        expected_result = '**kern\n=\n=\n=\n=\n=\n=\n=\n=\n*-\n'

        doc_1.add(doc_2)
        real_result = kp.dumps(doc_1)
        self.assertEqual(expected_result, real_result)

    def test_document_to_transposed_easy(self):
        doc, err = kp.load('resource_dir/legacy/base_tuplet_longer.krn')
        with open(Path('resource_dir/legacy/base_tuplet_longer_plus_octave.krn')) as f:
            expected_content = f.read()

        doc_transposed = doc.to_transposed('octave', 'up')

        real_content = kp.dumps(doc_transposed)

        self.assertEqual(expected_content, real_content)





class StoreCache:
    """
    A simple cache that stores the result of a callback function
    """
    def __init__(self):
        """
        Constructor
        """
        self.memory = {}

    def request(self, callback, request):
        """
        Request a value from the cache. If the value is not in the cache, it will be calculated by the callback function
        Args:
            callback (function): The callback function that will be called to calculate the value
            request (any): The request that will be passed to the callback function

        Returns (any): The value that was requested

        Examples:
            >>> def add_five(x):
            ...     return x + 5
            >>> store_cache = StoreCache()
            >>> store_cache.request(callback, 5)  # Call the callback function
            10
            >>> store_cache.request(callback, 5)  # Return the value from the cache, without calling the callback function
            10
        """
        if request in self.memory:
            return self.memory[request]
        else:
            result = callback(request)
            self.memory[request] = result
            return result




from __future__ import annotations
import sys

from .dyn_importer import DynSpineImporter
from .dynam_spine_importer import DynamSpineImporter
from .fing_spine_importer import FingSpineImporter
from .harm_spine_importer import HarmSpineImporter
from .kern_spine_importer import KernSpineImporter
from .mens_spine_importer import MensSpineImporter
from .mhxm_spine_importer import MxhmSpineImporter
from .root_spine_importer import RootSpineImporter
from .spine_importer import SpineImporter
from .text_spine_importer import TextSpineImporter
from .basic_spine_importer import BasicSpineImporter


def createImporter(spine_type: str) -> SpineImporter:
    if spine_type == '**mens':
        return MensSpineImporter()
    elif spine_type == '**kern':
        return KernSpineImporter()
    elif spine_type == '**text':
        return TextSpineImporter()
    elif spine_type == '**harm':
        return HarmSpineImporter()
    elif spine_type == '**mxhm':
        return MxhmSpineImporter()
    elif spine_type == '**root':
        return RootSpineImporter()
    elif spine_type == '**dyn':
        return DynSpineImporter()
    elif spine_type == '**dynam':
        return DynamSpineImporter()
    elif spine_type == '**fing':
        return FingSpineImporter()
    else:
        #print(f'Invalid spine header type found: {spine_type} '
        #      f'using a basic spine importer instead',
        #      file=sys.stderr)
        # TODO: Should we use a logger? Global variable for verbosity?

        return BasicSpineImporter()  # Only parse basic token categories




# Generated from kernSpineParser.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .kernSpineParser import kernSpineParser
else:
    from kernSpineParser import kernSpineParser

# This class defines a complete generic visitor for a parse tree produced by kernSpineParser.

class kernSpineParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by kernSpineParser#start.
    def visitStart(self, ctx:kernSpineParser.StartContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#field.
    def visitField(self, ctx:kernSpineParser.FieldContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#notes_rests_chords.
    def visitNotes_rests_chords(self, ctx:kernSpineParser.Notes_rests_chordsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#structural.
    def visitStructural(self, ctx:kernSpineParser.StructuralContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#contextual.
    def visitContextual(self, ctx:kernSpineParser.ContextualContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#signatures.
    def visitSignatures(self, ctx:kernSpineParser.SignaturesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#otherContextual.
    def visitOtherContextual(self, ctx:kernSpineParser.OtherContextualContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#empty.
    def visitEmpty(self, ctx:kernSpineParser.EmptyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#rest.
    def visitRest(self, ctx:kernSpineParser.RestContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#restChar_r.
    def visitRestChar_r(self, ctx:kernSpineParser.RestChar_rContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#restDecoration.
    def visitRestDecoration(self, ctx:kernSpineParser.RestDecorationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#chord.
    def visitChord(self, ctx:kernSpineParser.ChordContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#note.
    def visitNote(self, ctx:kernSpineParser.NoteContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#nonVisualTandemInterpretation.
    def visitNonVisualTandemInterpretation(self, ctx:kernSpineParser.NonVisualTandemInterpretationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#boundingBox.
    def visitBoundingBox(self, ctx:kernSpineParser.BoundingBoxContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#xywh.
    def visitXywh(self, ctx:kernSpineParser.XywhContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#x.
    def visitX(self, ctx:kernSpineParser.XContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#y.
    def visitY(self, ctx:kernSpineParser.YContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#w.
    def visitW(self, ctx:kernSpineParser.WContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#h.
    def visitH(self, ctx:kernSpineParser.HContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#pageNumber.
    def visitPageNumber(self, ctx:kernSpineParser.PageNumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#visualTandemInterpretation.
    def visitVisualTandemInterpretation(self, ctx:kernSpineParser.VisualTandemInterpretationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#associatedIDS.
    def visitAssociatedIDS(self, ctx:kernSpineParser.AssociatedIDSContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#placeHolder.
    def visitPlaceHolder(self, ctx:kernSpineParser.PlaceHolderContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#octaveShift.
    def visitOctaveShift(self, ctx:kernSpineParser.OctaveShiftContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#pianoHand.
    def visitPianoHand(self, ctx:kernSpineParser.PianoHandContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#tandemTuplet.
    def visitTandemTuplet(self, ctx:kernSpineParser.TandemTupletContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#tandemCue.
    def visitTandemCue(self, ctx:kernSpineParser.TandemCueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#tandemTremolo.
    def visitTandemTremolo(self, ctx:kernSpineParser.TandemTremoloContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#ossia.
    def visitOssia(self, ctx:kernSpineParser.OssiaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#rscale.
    def visitRscale(self, ctx:kernSpineParser.RscaleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#pedal.
    def visitPedal(self, ctx:kernSpineParser.PedalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#ela.
    def visitEla(self, ctx:kernSpineParser.ElaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#dynamics_position.
    def visitDynamics_position(self, ctx:kernSpineParser.Dynamics_positionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#sections.
    def visitSections(self, ctx:kernSpineParser.SectionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#sectionNames.
    def visitSectionNames(self, ctx:kernSpineParser.SectionNamesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#sectionName.
    def visitSectionName(self, ctx:kernSpineParser.SectionNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#transposition.
    def visitTransposition(self, ctx:kernSpineParser.TranspositionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#instrument.
    def visitInstrument(self, ctx:kernSpineParser.InstrumentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#instrumentTitle.
    def visitInstrumentTitle(self, ctx:kernSpineParser.InstrumentTitleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#number.
    def visitNumber(self, ctx:kernSpineParser.NumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#lowerCasePitch.
    def visitLowerCasePitch(self, ctx:kernSpineParser.LowerCasePitchContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#upperCasePitch.
    def visitUpperCasePitch(self, ctx:kernSpineParser.UpperCasePitchContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#pitchClass.
    def visitPitchClass(self, ctx:kernSpineParser.PitchClassContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#accomp.
    def visitAccomp(self, ctx:kernSpineParser.AccompContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#solo.
    def visitSolo(self, ctx:kernSpineParser.SoloContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#strophe.
    def visitStrophe(self, ctx:kernSpineParser.StropheContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#timebase.
    def visitTimebase(self, ctx:kernSpineParser.TimebaseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#part.
    def visitPart(self, ctx:kernSpineParser.PartContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#group.
    def visitGroup(self, ctx:kernSpineParser.GroupContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#staff.
    def visitStaff(self, ctx:kernSpineParser.StaffContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#clef.
    def visitClef(self, ctx:kernSpineParser.ClefContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#clefValue.
    def visitClefValue(self, ctx:kernSpineParser.ClefValueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#clefSign.
    def visitClefSign(self, ctx:kernSpineParser.ClefSignContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#clefLine.
    def visitClefLine(self, ctx:kernSpineParser.ClefLineContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#clefOctave.
    def visitClefOctave(self, ctx:kernSpineParser.ClefOctaveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#keySignature.
    def visitKeySignature(self, ctx:kernSpineParser.KeySignatureContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#keySignaturePitchClass.
    def visitKeySignaturePitchClass(self, ctx:kernSpineParser.KeySignaturePitchClassContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#keySignatureCancel.
    def visitKeySignatureCancel(self, ctx:kernSpineParser.KeySignatureCancelContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#keyCancel.
    def visitKeyCancel(self, ctx:kernSpineParser.KeyCancelContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#keyMode.
    def visitKeyMode(self, ctx:kernSpineParser.KeyModeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#key.
    def visitKey(self, ctx:kernSpineParser.KeyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#singleKey.
    def visitSingleKey(self, ctx:kernSpineParser.SingleKeyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#minorKey.
    def visitMinorKey(self, ctx:kernSpineParser.MinorKeyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#majorKey.
    def visitMajorKey(self, ctx:kernSpineParser.MajorKeyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#modal.
    def visitModal(self, ctx:kernSpineParser.ModalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#locrian.
    def visitLocrian(self, ctx:kernSpineParser.LocrianContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#ionian.
    def visitIonian(self, ctx:kernSpineParser.IonianContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#aeolian.
    def visitAeolian(self, ctx:kernSpineParser.AeolianContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#mixolydian.
    def visitMixolydian(self, ctx:kernSpineParser.MixolydianContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#lydian.
    def visitLydian(self, ctx:kernSpineParser.LydianContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#phrygian.
    def visitPhrygian(self, ctx:kernSpineParser.PhrygianContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#dorian.
    def visitDorian(self, ctx:kernSpineParser.DorianContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#timeSignature.
    def visitTimeSignature(self, ctx:kernSpineParser.TimeSignatureContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#numerator.
    def visitNumerator(self, ctx:kernSpineParser.NumeratorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#denominator.
    def visitDenominator(self, ctx:kernSpineParser.DenominatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#standardTimeSignature.
    def visitStandardTimeSignature(self, ctx:kernSpineParser.StandardTimeSignatureContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#additiveTimeSignature.
    def visitAdditiveTimeSignature(self, ctx:kernSpineParser.AdditiveTimeSignatureContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#mixedTimeSignature.
    def visitMixedTimeSignature(self, ctx:kernSpineParser.MixedTimeSignatureContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#alternatingTimeSignature.
    def visitAlternatingTimeSignature(self, ctx:kernSpineParser.AlternatingTimeSignatureContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#alternatingTimeSignatureItem.
    def visitAlternatingTimeSignatureItem(self, ctx:kernSpineParser.AlternatingTimeSignatureItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#interchangingTimeSignature.
    def visitInterchangingTimeSignature(self, ctx:kernSpineParser.InterchangingTimeSignatureContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#meterSymbol.
    def visitMeterSymbol(self, ctx:kernSpineParser.MeterSymbolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#modernMeterSymbolSign.
    def visitModernMeterSymbolSign(self, ctx:kernSpineParser.ModernMeterSymbolSignContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#mensuration.
    def visitMensuration(self, ctx:kernSpineParser.MensurationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#metronome.
    def visitMetronome(self, ctx:kernSpineParser.MetronomeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#nullInterpretation.
    def visitNullInterpretation(self, ctx:kernSpineParser.NullInterpretationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#barline.
    def visitBarline(self, ctx:kernSpineParser.BarlineContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#barLineType.
    def visitBarLineType(self, ctx:kernSpineParser.BarLineTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#restPosition.
    def visitRestPosition(self, ctx:kernSpineParser.RestPositionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#duration.
    def visitDuration(self, ctx:kernSpineParser.DurationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#fermata.
    def visitFermata(self, ctx:kernSpineParser.FermataContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#modernDuration.
    def visitModernDuration(self, ctx:kernSpineParser.ModernDurationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#augmentationDot.
    def visitAugmentationDot(self, ctx:kernSpineParser.AugmentationDotContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#alteration.
    def visitAlteration(self, ctx:kernSpineParser.AlterationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#staffChange.
    def visitStaffChange(self, ctx:kernSpineParser.StaffChangeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#chordSpace.
    def visitChordSpace(self, ctx:kernSpineParser.ChordSpaceContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#graceNote.
    def visitGraceNote(self, ctx:kernSpineParser.GraceNoteContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#appoggiatura.
    def visitAppoggiatura(self, ctx:kernSpineParser.AppoggiaturaContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#appoggiaturaMode.
    def visitAppoggiaturaMode(self, ctx:kernSpineParser.AppoggiaturaModeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#ligatureTie.
    def visitLigatureTie(self, ctx:kernSpineParser.LigatureTieContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#noteDecoration.
    def visitNoteDecoration(self, ctx:kernSpineParser.NoteDecorationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#noteDecorationCharX.
    def visitNoteDecorationCharX(self, ctx:kernSpineParser.NoteDecorationCharXContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#phrase.
    def visitPhrase(self, ctx:kernSpineParser.PhraseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#diatonicPitchAndOctave.
    def visitDiatonicPitchAndOctave(self, ctx:kernSpineParser.DiatonicPitchAndOctaveContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#trebleNotes.
    def visitTrebleNotes(self, ctx:kernSpineParser.TrebleNotesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#bassNotes.
    def visitBassNotes(self, ctx:kernSpineParser.BassNotesContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#accidental.
    def visitAccidental(self, ctx:kernSpineParser.AccidentalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#alterationDisplay.
    def visitAlterationDisplay(self, ctx:kernSpineParser.AlterationDisplayContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#turn.
    def visitTurn(self, ctx:kernSpineParser.TurnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#userAssignable.
    def visitUserAssignable(self, ctx:kernSpineParser.UserAssignableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#glissando.
    def visitGlissando(self, ctx:kernSpineParser.GlissandoContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#articulation.
    def visitArticulation(self, ctx:kernSpineParser.ArticulationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#accent.
    def visitAccent(self, ctx:kernSpineParser.AccentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#tenuto.
    def visitTenuto(self, ctx:kernSpineParser.TenutoContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#staccatissimo.
    def visitStaccatissimo(self, ctx:kernSpineParser.StaccatissimoContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#pizzicato.
    def visitPizzicato(self, ctx:kernSpineParser.PizzicatoContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#spiccato.
    def visitSpiccato(self, ctx:kernSpineParser.SpiccatoContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#staccato.
    def visitStaccato(self, ctx:kernSpineParser.StaccatoContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#editorialIntervention.
    def visitEditorialIntervention(self, ctx:kernSpineParser.EditorialInterventionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#slurStart.
    def visitSlurStart(self, ctx:kernSpineParser.SlurStartContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#ligatureTieStart.
    def visitLigatureTieStart(self, ctx:kernSpineParser.LigatureTieStartContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#tieContinue.
    def visitTieContinue(self, ctx:kernSpineParser.TieContinueContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#ligatureTieEnd.
    def visitLigatureTieEnd(self, ctx:kernSpineParser.LigatureTieEndContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#slurEnd.
    def visitSlurEnd(self, ctx:kernSpineParser.SlurEndContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#barLineCrossedNoteStart.
    def visitBarLineCrossedNoteStart(self, ctx:kernSpineParser.BarLineCrossedNoteStartContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#barLineCrossedNoteEnd.
    def visitBarLineCrossedNoteEnd(self, ctx:kernSpineParser.BarLineCrossedNoteEndContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#stem.
    def visitStem(self, ctx:kernSpineParser.StemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#beam.
    def visitBeam(self, ctx:kernSpineParser.BeamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#mordent.
    def visitMordent(self, ctx:kernSpineParser.MordentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#trill.
    def visitTrill(self, ctx:kernSpineParser.TrillContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by kernSpineParser#footnote.
    def visitFootnote(self, ctx:kernSpineParser.FootnoteContext):
        return self.visitChildren(ctx)



del kernSpineParser


"""
kernpy.core

=====

This module contains the core functionality of the `kernpy` package.
"""

from .tokens import *
from .document import *
from .importer import *
from .exporter import *
from .graphviz_exporter import  *
from .importer_factory import *
from .dyn_importer import *
from .dynam_spine_importer import *
from .fing_spine_importer import *
from .harm_spine_importer import *
from .kern_spine_importer import *
from .mens_spine_importer import *
from .root_spine_importer import *
from .text_spine_importer import *
from .mhxm_spine_importer import *
from .basic_spine_importer import *
from .generic import *
from .tokenizers import *
from .transposer import *
from .pitch_models import *
from .gkern import *


__all__ = [
    'Document',
    'TokenCategory',
    'Importer',
    'ExportOptions',
    'Exporter',
    'Encoding',
    'GraphvizExporter',
    'ekern_to_krn',
    'kern_to_ekern',
    'get_kern_from_ekern',
    'Encoding',
    'Tokenizer',
    'KernTokenizer',
    'EkernTokenizer',
    'BekernTokenizer',
    'BkernTokenizer',
    'AEKernTokenizer',
    'AKernTokenizer',
    'TokenizerFactory',
    'Token',
    'KernTokenizer',
    'BEKERN_CATEGORIES',
    'DynSpineImporter',
    'DynamSpineImporter',
    'FingSpineImporter',
    'HarmSpineImporter',
    'KernSpineImporter',
    'MensSpineImporter',
    'RootSpineImporter',
    'TextSpineImporter',
    'MxhmSpineImporter',
    'BasicSpineImporter',
    'SpineOperationToken',
    'PitchRest',
    'Duration',
    'DurationClassical',
    'DurationMensural',
    'read',
    'create',
    'export',
    'store',
    'store_graph',
    'transposer',
    'get_spine_types',
    'createImporter',
    'TokenCategoryHierarchyMapper',
    'TOKEN_SEPARATOR',
    'DECORATION_SEPARATOR',
    'Subtoken',
    'AbstractToken',
    'SimpleToken',
    'ComplexToken',
    'CompoundToken',
    'NoteRestToken',
    'HeaderToken',
    'HeaderTokenGenerator',
    'NotationEncoding',
    'AgnosticPitch',
    'PitchExporter',
    'PitchExporterFactory',
    'HumdrumPitchExporter',
    'AmericanPitchExporter',
    'PitchImporter',
    'PitchImporterFactory',
    'HumdrumPitchImporter',
    'AmericanPitchImporter',
    'Direction',
    'Intervals',
    'IntervalsByName',
    'transpose',
    'transpose_agnostics',
    'transpose_encoding_to_agnostic',
    'transpose_agnostic_to_encoding',
    'PositionInStaff',
    'distance',
    'agnostic_distance',
    'PitchPositionReferenceSystem',
    'Clef',
    'GClef',
    'F3Clef',
    'F4Clef',
    'C1Clef',
    'C2Clef',
    'C3Clef',
    'C4Clef',
    'ClefFactory',
    'AVAILABLE_INTERVALS',
    'GKernExporter',
    'Staff',
    'pitch_to_gkern_string',
    'gkern_to_g_clef_pitch',
]




from __future__ import annotations

from abc import ABC, abstractmethod

from antlr4 import CommonTokenStream, ParseTreeWalker
from antlr4.InputStream import InputStream


class BaseANTLRImporter(ABC):
    def __init__(self, input_string):
        char_stream = InputStream(input_string)
        self.lexer = self.createLexer(char_stream)
        token_stream = CommonTokenStream(self.lexer)
        self.parser = self.createParser(token_stream)

    @abstractmethod
    def createLexer(self, charStream):
        pass

    @abstractmethod
    def createParser(self, tokenStream):
        pass

    @abstractmethod
    def startRule(self):
        pass


class BaseANTLRListenerImporter(BaseANTLRImporter):
    def __init__(self, input_string):
        super().__init__(input_string)
        self.listener = self.createListener()

    def start(self):
        tree = self.startRule()
        walker = ParseTreeWalker()
        walker.walk(self.listener, tree)
        #ParseTreeWalker.DEFAULT.walk(walker, tree)

    @abstractmethod
    def createListener(self):
        pass


class BaseANTLRVisitorImporter(BaseANTLRImporter):
    def __init__(self, input):
        super().__init__(input)
        self.visitor = self.createVisitor()

    def start(self):
        tree_context = self.startRule()
        self.visitStart(tree_context)

    @abstractmethod
    def createVisitor(self):
        pass

    @abstractmethod
    def visitStart(self, start_context):
        pass

